import numpy as np
import numpy.testing as npt
import pytest

import spharpy
from spharpy.samplings.coordinates import Coordinates, SamplingSphere
from spharpy.samplings import sph2cart, cart2sph, cart2latlon


def test_coordinates_init():
    coords = Coordinates()
    assert isinstance(coords, Coordinates)


def test_coordinates_init_val():

    coords = Coordinates(1, 0, 0)
    assert isinstance(coords, Coordinates)


def test_coordinates_init_incomplete():
    x = [1, 2]
    y = 1
    z = 1
    with pytest.raises(ValueError):
        Coordinates(x, y, z)
        pytest.fail("Input arrays need to have same dimensions.")


def test_coordinates_init_from_cartesian():
    x = 1
    y = 0
    z = 0
    coords = Coordinates.from_cartesian(x, y, z)
    npt.assert_allclose(coords._x, x)
    npt.assert_allclose(coords._y, y)
    npt.assert_allclose(coords._z, z)


def test_coordinates_init_from_spherical():
    x = 1
    y = 0
    z = 0
    rad, theta, phi = cart2sph(x, y, z)
    coords = Coordinates.from_spherical(rad, theta, phi)
    # use atol here because of numerical rounding issues introduced in
    # the coordinate conversion
    npt.assert_allclose(coords._x, x, atol=1e-15)
    npt.assert_allclose(coords._y, y, atol=1e-15)
    npt.assert_allclose(coords._z, z, atol=1e-15)

def test_coordinates_init_from_array_spherical():
    rad = [1., 1., 1., 1.]
    ele = [np.pi/2, np.pi/2, 0, np.pi/2]
    azi = [0, np.pi/2, 0, np.pi/4]

    points = np.array([rad, ele, azi])
    coords = Coordinates.from_array(points, coordinate_system='spherical')

    npt.assert_allclose(coords.radius, rad, atol=1e-15)
    npt.assert_allclose(coords.elevation, ele, atol=1e-15)
    npt.assert_allclose(coords.azimuth, azi, atol=1e-15)

def test_coordinates_init_from_array_cartesian():
    x = [1, 0, 0, 0]
    y = [0, 1, 0, 0]
    z = [0, 0, 1, 0]

    points = np.array([x, y, z])
    coords = Coordinates.from_array(points)

    npt.assert_allclose(coords._x, x, atol=1e-15)
    npt.assert_allclose(coords._y, y, atol=1e-15)
    npt.assert_allclose(coords._z, z, atol=1e-15)

def test_getter_x():
    x = np.array([1, 0], dtype=float)
    coords = Coordinates()
    coords._x = x
    npt.assert_allclose(coords.x, x)

def test_getter_y():
    y = np.array([1, 0], dtype=float)
    coords = Coordinates()
    coords._y = y
    npt.assert_allclose(coords.y, y)

def test_getter_z():
    z = np.array([1, 0], dtype=float)
    coords = Coordinates()
    coords._z = z
    npt.assert_allclose(coords.z, z)

def test_setter_x():
    value = np.array([1.0, 1], dtype=float)
    coords = Coordinates()
    coords.x = value
    npt.assert_allclose(value, coords._x)

def test_setter_y():
    value = np.array([1.0, 1], dtype=float)
    coords = Coordinates()
    coords.y = value
    npt.assert_allclose(value, coords._y)

def test_setter_z():
    value = np.array([1.0, 1], dtype=float)
    coords = Coordinates()
    coords.z = value
    npt.assert_allclose(value, coords._z)

def test_getter_ele():
    value = np.pi/2
    coords = Coordinates()
    coords.z = 0
    coords.y = 0
    coords.x = 1
    npt.assert_allclose(coords.elevation, value)

def test_getter_radius():
    value = 1
    coords = Coordinates()
    coords.z = 0
    coords.y = 1
    coords.x = 0
    npt.assert_allclose(coords.radius, value)

def test_getter_azi():
    azi = np.pi/2
    coords = Coordinates()
    coords.z = 0
    coords.y = 1
    coords.x = 0
    npt.assert_allclose(coords.azimuth, azi)

def test_setter_rad():
    eps = np.spacing(1)
    rad = 0.5
    x = 0.5
    y = 0
    z = 0
    coords = Coordinates(1, 0, 0)
    coords.radius = rad
    npt.assert_allclose(coords._x, x, atol=eps)
    npt.assert_allclose(coords._y, y, atol=eps)
    npt.assert_allclose(coords._z, z, atol=eps)

def test_setter_ele():
    eps = np.spacing(1)
    ele = 0
    x = 0
    y = 0
    z = 1
    coords = Coordinates(1, 0, 0)
    coords.elevation = ele
    npt.assert_allclose(coords._x, x, atol=eps)
    npt.assert_allclose(coords._y, y, atol=eps)
    npt.assert_allclose(coords._z, z, atol=eps)


def test_setter_azi():
    eps = np.spacing(1)
    azi = np.pi/2
    x = 0
    y = 1
    z = 0
    coords = Coordinates(1, 0, 0)
    coords.azimuth = azi
    npt.assert_allclose(coords._x, x, atol=eps)
    npt.assert_allclose(coords._y, y, atol=eps)
    npt.assert_allclose(coords._z, z, atol=eps)

def test_getter_latitude():
    x = 1
    y = 0
    z = 0.5

    height, lat, lon = cart2latlon(x, y, z)
    coords = Coordinates(x, y, z)
    npt.assert_allclose(coords.latitude, lat)

def test_getter_longitude():
    x = 1
    y = 0
    z = 0.5

    height, lat, lon = cart2latlon(x, y, z)
    coords = Coordinates(x, y, z)
    npt.assert_allclose(coords.longitude, lon)

def test_getter_cartesian():
    x = [1, 0, 0, 0]
    y = [0, 1, 0, 0]
    z = [0, 0, 1, 0]

    coords = Coordinates(x, y, z)
    ref = np.vstack((x, y, z))
    npt.assert_allclose(coords.cartesian, ref)


def test_setter_cartesian():
    x = np.array([1, 0, 0, 0])
    y = np.array([0, 1, 0, 0])
    z = np.array([0, 0, 1, 0])
    cart = np.vstack((x, y, z))
    coords = Coordinates()
    coords.cartesian = cart
    npt.assert_allclose(coords.cartesian, cart)


def test_getter_spherical():
    x = np.array([1, 0, 0, 1], dtype=float)
    y = np.array([0, 1, 0, 1], dtype=float)
    z = np.array([0, 0, 1, 1], dtype=float)

    rad, theta, phi = cart2sph(x, y, z)

    coords = Coordinates(x, y, z)
    ref = np.vstack((rad, theta, phi))
    npt.assert_allclose(coords.spherical, ref)


def test_setter_spherical():
    eps = np.spacing(1)
    x = np.array([1, 0, 0, 1], dtype=float)
    y = np.array([0, 1, 0, 1], dtype=float)
    z = np.array([0, 0, 1, 1], dtype=float)
    rad, theta, phi = cart2sph(x, y, z)
    spherial = np.vstack((rad, theta, phi))
    coords = Coordinates()
    coords.spherical = spherial
    npt.assert_allclose(coords._x, x, atol=eps)
    npt.assert_allclose(coords._y, y, atol=eps)
    npt.assert_allclose(coords._z, z, atol=eps)


def test_n_points():
    coords = Coordinates([1, 0], [1, 1], [0, 1])
    assert coords.n_points == 2


def test_find_nearest():
    coords = Coordinates([1, 0], [1, 1], [0, 1])
    point = Coordinates(1, 1, 0)

    dist, idx = coords.find_nearest_point(point)

    assert idx == 0


def test_len():
    coords = Coordinates([1, 0], [1, 1], [0, 1])
    assert len(coords) == 2


def test_getitem():
    coords = Coordinates([1, 0], [1, 1], [0, 1])
    getcoords = coords[0]
    npt.assert_allclose(np.squeeze(getcoords.cartesian), np.array([1, 1, 0]))


def test_setitem():
    coords = Coordinates([0, 0], [1, 1], [0, 1])
    setcoords = Coordinates(1, 1, 0)
    coords[0] = setcoords
    npt.assert_allclose(np.squeeze(coords.cartesian),
                        np.array([[1, 0], [1, 1], [0, 1]]))


def test_sampling_sphere_init():
    sampling = SamplingSphere()
    assert isinstance(sampling, SamplingSphere)


def test_sampling_sphere_init_value():
    sampling = SamplingSphere(1, 0, 0, 0)
    assert isinstance(sampling, SamplingSphere)


def sampling_cube():
    """Helper function returning a cube sampling"""
    x = [1, -1, 0, 0, 0, 0]
    y = [0, 0, 1, -1, 0, 0]
    z = [0, 0, 0, 0, 1, -1]

    return x, y, z

def test_getter_n_max():
    x, y, z = sampling_cube()
    n_max = 1
    sampling = SamplingSphere(x, y, z, n_max)

    assert sampling.n_max == n_max

def test_setter_n_max():
    x, y, z = sampling_cube()
    n_max = 1
    sampling = SamplingSphere(x, y, z, 0)

    sampling.n_max = n_max
    assert sampling._n_max == n_max


def test_merge():
    s1 = Coordinates(1, 0, 0)
    s2 = Coordinates(0, 2, 0)

    s1.merge(s2)

    truth = np.array([[1, 0], [0, 2], [0, 0]])
    npt.assert_allclose(truth, s1.cartesian)



import matplotlib.pyplot as plt
import numpy as np
import matplotlib.cm as cmap

import spharpy
import cartopy.crs as ccrs

sampling = spharpy.samplings.equalarea(20, condition_num=np.inf)
# sampling = spharpy.samplings.hyperinterpolation(15)
Y = spharpy.spherical.spherical_harmonic_basis_real(15, sampling)

# plt.figure()
# spharpy.plot.contour_map(sampling, (Y[:, spharpy.spherical.nm2acn(4, 2)]), cmap=cmap.Spectral_r)
#
# plt.figure()
# spharpy.plot.contour_map(sampling, (Y[:, 4]), cmap=cmap.RdBu_r, projection=ccrs.Mollweide())


# plt.figure(figsize=(5, 5))
# plot = spharpy.plot.balloon_wireframe(sampling, Y[:, 1], phase=False, cmap=cmap.RdBu_r)



# plt.figure(figsize=(5,5))
# plot = spharpy.plot.balloon(sampling, Y[:, 1], phase=False, cmap=cmap.RdBu_r)


# plt.figure(figsize=(5,5))
# plot = spharpy.plot.balloon_wireframe(sampling, Y[:, 7], phase=False, cmap=cmap.RdBu_r, colorbar=False)


# spharpy.plot.contour_map(sampling, Y[:, 7], cmap=cmap.RdBu_r, projection=ccrs.Sinusoidal())
plt.figure()
spharpy.plot.contour(sampling, Y[:, 7], cmap=cmap.RdBu_r)

plt.figure()
spharpy.plot.contour_map(sampling, Y[:, 7], cmap=cmap.RdBu_r)

plt.figure()
spharpy.plot.pcolor_map(sampling, Y[:, 7], cmap=cmap.RdBu_r, limits=[-0.3, 1])

# plt.figure()
# spharpy.plot.contour_map(sampling, Y[:, 2], cmap=cmap.RdBu_r, projection=ccrs.Sinusoidal())
#
# plt.figure()
# spharpy.plot.contour_map(sampling, Y[:, 2], cmap=cmap.RdBu_r, projection=ccrs.PlateCarree())



#
# plt.figure(figsize=(5,5))
# spharpy.plot.balloon_interp(sampling, Y[:, 1])

#
# Y = spharpy.spherical.spherical_harmonic_basis(15, sampling)
#
# plt.figure(figsize=(5,5))
# plot = spharpy.plot.balloon_wireframe(sampling, Y[:, 1], phase=False,colorbar=False)
#
# plt.figure(figsize=(5,5))
# plot = spharpy.plot.balloon(sampling, Y[:, 1], phase=True)
#
#



import matplotlib as mpl
# Use Agg backend to prevent matplotlib from creating interactive plots. This
# has to be set before the importing matplotlib.pyplot. Use switch backend in
# case the wrong backend has already been set.
mpl.use('agg')
import matplotlib.pyplot as plt
plt.switch_backend('agg')

import spharpy
import numpy as np
import pytest
from spharpy import plot


def test_balloon_plot_abs():
    coords = spharpy.samplings.icosahedron()
    data = np.cos(coords.azimuth)*np.sin(coords.elevation)
    spharpy.plot.balloon(coords, data, show=False)


def test_balloon_plot_phase():
    coords = spharpy.samplings.hyperinterpolation(10)
    data = np.cos(coords.azimuth)*np.sin(coords.elevation)
    spharpy.plot.balloon(coords, data, phase=True, show=False)


def test_contour_plot():
    coords = spharpy.samplings.hyperinterpolation(10)
    data = np.cos(coords.azimuth)*np.sin(coords.elevation)
    spharpy.plot.contour(coords, np.abs(data), show=False)


def test_scatter():
    """Test if the plot executes without raising an exception
    """
    coords = spharpy.samplings.hyperinterpolation(10)
    spharpy.plot.scatter(coords)

    # test of auto detection of axes works
    ax = plt.axes(projection='3d')
    spharpy.plot.scatter(coords)

    # explicitly pass axes
    spharpy.plot.scatter(coords, ax=ax)

    # pass axes with wrong projection
    ax = plt.axes()
    with pytest.raises(ValueError, match='3d'):
        spharpy.plot.scatter(coords, ax=ax)

    # current axis with wrong projection
    with pytest.raises(ValueError, match='3d'):
        spharpy.plot.scatter(coords)

    plt.close('all')


def test_pcolor_map():
    """Test if the plot executes without raising an exception
    """
    coords = spharpy.samplings.hyperinterpolation(10)
    data = np.cos(coords.azimuth)*np.sin(coords.elevation)
    plot.pcolor_map(coords, data)

    # test of auto detection of axes works
    ax = plt.axes(projection='mollweide')
    plot.pcolor_map(coords, data)

    # explicitly pass axes
    plot.pcolor_map(coords, data, ax=ax)

    # pass axes with wrong projection
    ax = plt.axes()
    with pytest.raises(ValueError, match='Projection does not match'):
        plot.pcolor_map(coords, data, ax=ax)

    plt.close('all')


def test_contour_map():
    """Test if the plot executes without raising an exception
    """
    coords = spharpy.samplings.hyperinterpolation(10)
    data = np.cos(coords.azimuth)*np.sin(coords.elevation)
    plot.contour_map(coords, data)

    # test of auto detection of axes works
    ax = plt.axes(projection='mollweide')
    plot.contour_map(coords, data)

    # explicitly pass axes
    plot.contour_map(coords, data, ax=ax)

    # pass axes with wrong projection
    ax = plt.axes()
    with pytest.raises(ValueError, match='Projection does not match'):
        plot.contour_map(coords, data, ax=ax)

    plt.close('all')


def test_contour():
    """Test if the plot executes without raising an exception
    """
    coords = spharpy.samplings.hyperinterpolation(10)
    data = np.cos(coords.azimuth)*np.sin(coords.elevation)
    plot.contour_map(coords, data)

    # test of auto detection of axes works
    ax = plt.axes()
    plot.contour(coords, data)

    # explicitly pass axes
    plot.contour(coords, data, ax=ax)

    plt.close('all')


def test_plot_voronoi_sphere():
    """Test if the plot executes without raising an exception
    """
    coords = spharpy.samplings.hyperinterpolation(4)
    plot.voronoi_cells_sphere(coords)

    # test of auto detection of axes works
    ax = plt.axes(projection='3d')
    plot.voronoi_cells_sphere(coords)

    # explicitly pass axes
    plot.voronoi_cells_sphere(coords, ax=ax)

    # pass axes with wrong projection
    ax = plt.axes()
    with pytest.raises(ValueError, match='3d'):
        plot.voronoi_cells_sphere(coords, ax=ax)

    # current axis with wrong projection
    with pytest.raises(ValueError, match='3d'):
        plot.voronoi_cells_sphere(coords)

    plt.close('all')


def test_pcolor_sphere():
    coords = spharpy.samplings.hyperinterpolation(10)
    data = np.cos(coords.azimuth)*np.sin(coords.elevation)
    plot.pcolor_sphere(coords, data, colorbar=True)

    # test of auto detection of axes works
    ax = plt.axes(projection='3d')
    plot.pcolor_sphere(coords, data)

    # explicitly pass axes
    plot.pcolor_sphere(coords, data, ax=ax)

    # pass axes with wrong projection
    ax = plt.axes()
    with pytest.raises(ValueError, match="'3d'"):
        plot.pcolor_sphere(coords, data, ax=ax)

    plt.close('all')


def test_balloon_wireframe():
    coords = spharpy.samplings.hyperinterpolation(10)
    data = np.cos(coords.azimuth)*np.sin(coords.elevation)
    plot.balloon_wireframe(coords, data)

    # test of auto detection of axes works
    ax = plt.axes(projection='3d')
    plot.balloon_wireframe(coords, data)

    # explicitly pass axes
    plot.balloon_wireframe(coords, data, ax=ax)

    # pass axes with wrong projection
    ax = plt.axes()
    with pytest.raises(ValueError, match="'3d'"):
        plot.balloon_wireframe(coords, data, ax=ax)

    plt.close('all')


def test_balloon():
    coords = spharpy.samplings.hyperinterpolation(10)
    data = np.cos(coords.azimuth)*np.sin(coords.elevation)
    plot.balloon(coords, data)

    # test of auto detection of axes works
    ax = plt.axes(projection='3d')
    plot.balloon(coords, data)

    # explicitly pass axes
    plot.balloon(coords, data, ax=ax)

    # pass axes with wrong projection
    ax = plt.axes()
    with pytest.raises(ValueError, match="'3d'"):
        plot.balloon(coords, data, ax=ax)

    plt.close('all')



"""
Tests for acn and order, degree indexing functions
"""

import sys
from filehandling import read_matrix_from_mat

sys.path.append('./')

import pytest
import spharpy.spherical as sh
import numpy as np

def test_acn2nm_single_val():
    n,m = sh.acn2nm(0)
    assert n == 0
    assert m == 0

    n, m = sh.acn2nm(2)
    assert n == 1
    assert m == 0

    n, m = sh.acn2nm(1)
    assert n == 1
    assert m == -1


def test_nm2acn_single_val():
    acn = sh.nm2acn(0, 0)
    assert acn == 0

    acn = sh.nm2acn(1, 0)
    assert acn == 2

    acn = sh.nm2acn(1, -1)
    assert acn == 1

def test_acn2nm_array():
    n_ref = np.array([0, 1, 1, 1])
    m_ref = np.array([0, -1, 0, 1])

    n_max = 1
    n_sh = (n_max + 1)**2
    acn = np.arange(0, n_sh)
    n, m = sh.acn2nm(acn)

    np.testing.assert_equal(n, n_ref)
    np.testing.assert_equal(m, m_ref)

def test_nm2acn_array():
    n = np.array([0, 1, 1, 1])
    m = np.array([0, -1, 0, 1])

    n_max = 1
    n_sh = (n_max + 1)**2
    acn_ref = np.arange(0, n_sh)

    acn = sh.nm2acn(n, m)

    np.testing.assert_equal(acn, acn_ref)



""" Tests for spatial sampling functions """

import numpy as np
import pytest

import spharpy.samplings as samplings
from spharpy.samplings.coordinates import Coordinates, SamplingSphere


def test_cube_equidistant():
    n_points = 3
    coords = samplings.cube_equidistant(n_points)
    x = np.tile(np.array([-1, -1, -1, 0, 0, 0, 1, 1, 1]), 3)
    y = np.hstack((np.ones(9) * -1, np.zeros(9), np.ones(9)))
    z = np.tile(np.array([-1, 0, 1]), 9)
    np.testing.assert_allclose(x, coords.x)
    np.testing.assert_allclose(y, coords.y)
    np.testing.assert_allclose(z, coords.z)


def test_hyperinterpolation():
    n_max = 1
    sampling = samplings.hyperinterpolation(n_max)
    assert sampling.radius.size == (n_max+1)**2


def test_spherical_t_design_const_e():
    order = 2
    coords = samplings.spherical_t_design(
        order, criterion='const_energy')
    assert isinstance(coords, SamplingSphere)


def test_spherical_t_design_const_angle():
    order = 2
    coords = samplings.spherical_t_design(
        order, criterion='const_angular_spread')
    assert isinstance(coords, SamplingSphere)


def test_spherical_t_design_invalid():
    order = 2
    with pytest.raises(ValueError, match='Invalid design'):
        samplings.spherical_t_design(order, criterion='bla')


def test_dodecahedron():
    sampling = samplings.dodecahedron()
    assert isinstance(sampling, SamplingSphere)


def test_icosahedron():
    sampling = samplings.icosahedron()
    assert isinstance(sampling, SamplingSphere)


def test_equiangular():
    n_max = 1
    sampling = samplings.equiangular(n_max)
    assert isinstance(sampling, SamplingSphere)


def test_gaussian():
    n_max = 1
    sampling = samplings.gaussian(n_max)
    assert isinstance(sampling, SamplingSphere)


def test_em32():
    sampling = samplings.eigenmike_em32()
    assert isinstance(sampling, SamplingSphere)


def test_icosahedron_ke4():
    sampling = samplings.icosahedron_ke4()
    assert isinstance(sampling, SamplingSphere)


def test_equalarea():
    sampling = samplings.equalarea(2)
    assert isinstance(sampling, SamplingSphere)


def test_spiral_points():
    sampling = samplings.spiral_points(2)
    assert isinstance(sampling, SamplingSphere)



"""
Tests for spherical harmonic basis and related functions
"""

import sys
from filehandling import read_2d_matrix_from_csv

sys.path.append('./')

from unittest.mock import patch


import spharpy.spherical as sh
import spharpy.samplings as samplings
from spharpy.samplings import Coordinates
import numpy as np
import numpy.testing as npt


def test_spherical_harmonic():
    Nmax = 1
    theta = np.array([np.pi/2, np.pi/2, 0], dtype=float)
    phi = np.array([0, np.pi/2, 0], dtype=float)

    rad = np.ones(3, dtype=float)
    coords = Coordinates.from_spherical(rad, theta, phi)

    Y = np.array([[2.820947917738781e-01 + 0.000000000000000e+00j, 3.454941494713355e-01 + 0.000000000000000e+00j, 2.991827511286337e-17 + 0.000000000000000e+00j, -3.454941494713355e-01 + 0.000000000000000e+00j],
                  [2.820947917738781e-01 + 0.000000000000000e+00j, 2.115541521371041e-17 - 3.454941494713355e-01j, 2.991827511286337e-17 + 0.000000000000000e+00j, -2.115541521371041e-17 - 3.454941494713355e-01j],
                  [2.820947917738781e-01 + 0.000000000000000e+00j, 0.000000000000000e+00 + 0.000000000000000e+00j, 4.886025119029199e-01 + 0.000000000000000e+00j, 0.000000000000000e+00 + 0.000000000000000e+00j]], dtype=complex)

    basis = sh.spherical_harmonic_basis(Nmax, coords)

    np.testing.assert_allclose(Y, basis, atol=1e-13)


def test_spherical_harmonic_n10():
    Nmax = 10
    theta = np.array([np.pi/2, np.pi/2, 0], dtype=float)
    phi = np.array([0, np.pi/2, 0], dtype=float)
    n_points = len(theta)

    with patch.multiple(
            Coordinates,
            azimuth=phi,
            elevation=theta,
            n_points=n_points) as patched_vals:
        coords = Coordinates()

        Y = np.genfromtxt(
            './tests/data/sh_basis_cplx_n10.csv',
            delimiter=',',
            dtype=complex)

        basis = sh.spherical_harmonic_basis(Nmax, coords)

        np.testing.assert_allclose(Y, basis, atol=1e-13)


def test_spherical_harmonics_real():
    n_max = 10
    theta = np.array([np.pi/2, np.pi/2, 0, np.pi/2], dtype=float)
    phi = np.array([0, np.pi/2, 0, np.pi/4], dtype=float)
    rad = np.ones(4)
    coords = Coordinates.from_spherical(rad, theta, phi)

    reference = read_2d_matrix_from_csv('./tests/data/sh_basis_real.csv')
    basis = sh.spherical_harmonic_basis_real(n_max, coords)
    np.testing.assert_allclose(basis, reference, atol=1e-13)


def test_orthogonality():
    """
    Check if the orthonormality condition of the spherical harmonics is fulfilled
    """
    n_max = 82
    theta = np.array([np.pi/2, np.pi/2, 0, np.pi/2], dtype=float)
    phi = np.array([0, np.pi/2, 0, np.pi/4], dtype=float)
    n_points = phi.size
    n_points = len(theta)

    with patch.multiple(
            Coordinates,
            azimuth=phi,
            elevation=theta,
            n_points=n_points) as patched_vals:
        coords = Coordinates()
        basis = sh.spherical_harmonic_basis(n_max, coords)

        inner = (basis @ np.conjugate(basis.T))
        fact = 4 * np.pi / (n_max + 1) ** 2
        orth = np.diagonal(fact * inner)
        np.testing.assert_allclose(
            orth, np.ones(n_points, dtype=complex), rtol=1e-14)


def test_orthogonality_real():
    """
    Check if the orthonormality condition of the reavl valued spherical harmonics is fulfilled
    """
    n_max = 82
    theta = np.array([np.pi / 2, np.pi / 2, 0, np.pi / 2], dtype='double')
    phi = np.array([0, np.pi / 2, 0, np.pi / 4], dtype='double')
    n_points = phi.size

    with patch.multiple(
            Coordinates,
            azimuth=phi,
            elevation=theta,
            n_points=n_points) as patched_vals:
        coords = Coordinates()
        basis = sh.spherical_harmonic_basis_real(n_max, coords)

        inner = (basis @ np.conjugate(basis.T))
        fact = 4 * np.pi / (n_max + 1) ** 2
        orth = np.diagonal(fact * inner)
        np.testing.assert_allclose(orth, np.ones(n_points), rtol=1e-15, atol=1e-10)


def test_spherical_harmonic_basis_gradient():
    n_max = 15
    theta = np.array([np.pi/2, np.pi/2, 0, np.pi/2, np.pi/4])
    phi = np.array([0, np.pi/2, 0, np.pi/4, np.pi/4])
    n_points = np.size(theta)

    with patch.multiple(
            Coordinates,
            azimuth=phi,
            elevation=theta,
            n_points=n_points):
        coords = Coordinates()

        grad_ele, grad_azi = \
            sh.spherical_harmonic_basis_gradient(n_max, coords)

        desire_ele = np.genfromtxt(
            './tests/data/Y_grad_ele.csv',
            dtype=complex,
            delimiter=',')
        npt.assert_allclose(grad_ele, desire_ele, rtol=1e-10, atol=1e-10)

        desire_azi = np.genfromtxt(
            './tests/data/Y_grad_azi.csv',
            dtype=complex,
            delimiter=',')
        npt.assert_allclose(grad_azi, desire_azi, rtol=1e-10, atol=1e-10)


def test_spherical_harmonic_basis_gradient_real():
    n_max = 15
    theta = np.array([np.pi/2, np.pi/2, 0, np.pi/2, np.pi/4])
    phi = np.array([0, np.pi/2, 0, np.pi/4, np.pi/4])
    n_points = np.size(theta)

    with patch.multiple(
            Coordinates,
            azimuth=phi,
            elevation=theta,
            n_points=n_points):
        coords = Coordinates()

        grad_ele, grad_azi = \
            sh.spherical_harmonic_basis_gradient_real(n_max, coords)

        desire_ele = np.genfromtxt(
            './tests/data/Y_grad_real_ele.csv',
            dtype=complex,
            delimiter=',')
        npt.assert_allclose(grad_ele, desire_ele, rtol=1e-10, atol=1e-10)

        desire_azi = np.genfromtxt(
            './tests/data/Y_grad_real_azi.csv',
            dtype=complex,
            delimiter=',')
        npt.assert_allclose(grad_azi, desire_azi, rtol=1e-10, atol=1e-10)



"""
Commonly used channel indexing functions used in Ambisonics
"""

import numpy as np
import spharpy


def sid(n_max):
    """Calculates the SID indices up to spherical harmonic order n_max.
    TODO: Add citation Daniel

    Parameters
    ----------
    n_max : int

    Returns
    -------
    sid : numpy array

    """
    n_sh = (n_max+1)**2
    sid_n = sph_identity_matrix(n_max, 'n-nm').T @ np.arange(0, n_max+1)
    sid_m = np.zeros(n_sh, dtype=int)
    idx_n = 0
    for n in range(1, n_max+1):
        for m in range(1, n+1):
            sid_m[idx_n + 2*m-1] = n-m+1
            sid_m[idx_n + 2*m] = -(n-m+1)
        sid_m[idx_n + 2*n + 1] = 0
        idx_n += 2*n+1

    return sid_n, sid_m


def sid2acn(n_max):
    """Convert from SID channel indexing as proposed by Daniel in
    TODO: add citation
    Returns the indices to achieve a corresponding linear acn indexing.

    Parameters
    ----------
    sid : numpy array

    Returns
    -------
    acn : numpy array

    """
    sid_n, sid_m = spharpy.indexing.sid(n_max)
    linear_sid = spharpy.spherical.nm2acn(sid_n, sid_m)
    return np.argsort(linear_sid)


def sph_identity_matrix(n_max, type='n-nm'):
    """Calculate a spherical harmonic identity matrix.
    TODO: Implement the other identity matrices

    Parameters
    ----------
    n_max : TODO
    type : TODO, optional

    Returns
    -------
    identity_matrix : numpy array

    """
    n_sh = (n_max+1)**2

    if type != 'n-nm':
        raise NotImplementedError

    identity_matrix = np.zeros((n_max+1, n_sh), dtype=int)
    # linear_n0 = np.cumsum(np.arange(0, 2*(n_max+1), 2))

    for n in range(n_max+1):
        m = np.arange(-n, n+1)
        linear_nm = spharpy.spherical.nm2acn(np.tile(n, m.shape), m)
        identity_matrix[n, linear_nm] = 1

    return identity_matrix



"""
File handling helper functions
"""
import numpy as np
from scipy.io import loadmat


def read_2d_matrix_from_csv(filename, dtype='double'):
    """
    Read 2d matrix from csv file
    """
    return np.genfromtxt(open(filename, "rb"), delimiter=",", dtype=dtype)


def read_matrix_from_mat(filename):
    """
    Read matrix from .mat file as numpy ndarray
    """
    return loadmat(filename)['matrix']



from .samplings import (
    cube_equidistant,
    hyperinterpolation,
    spherical_t_design,
    dodecahedron,
    icosahedron,
    icosahedron_ke4,
    equiangular,
    gaussian,
    eigenmike_em32,
    equalarea,
    spiral_points)

from .helpers import (
    sph2cart,
    cart2sph,
    cart2latlon,
    latlon2cart,
    spherical_voronoi,
    calculate_sampling_weights
)

from .coordinates import Coordinates, SamplingSphere

from .interior import interior_stabilization_points


__all__ = [
    'cube_equidistant',
    'hyperinterpolation',
    'spherical_t_design',
    'dodecahedron',
    'icosahedron',
    'icosahedron_ke4',
    'equiangular',
    'gaussian',
    'eigenmike_em32',
    'equalarea',
    'spiral_points',
    'sph2cart',
    'cart2sph',
    'cart2latlon',
    'latlon2cart',
    'spherical_voronoi',
    'calculate_sampling_weights',
    'Coordinates',
    'SamplingSphere',
    'interior_stabilization_points',
]



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import spharpy
from matplotlib.collections import LineCollection

phi = np.linspace(0, 2*np.pi, 512)
func = np.exp(1j*phi)

fig, axs = plt.subplots(1, 2, figsize=(12, 4))

points = np.array([np.real(func), np.imag(func)]).T.reshape(-1, 1, 2)
segments = np.concatenate([points[:-1], points[1:]], axis=1)

norm = plt.Normalize(phi.min(), phi.max())
lc = LineCollection(segments, cmap=cm.twilight, norm=norm)
# # Set the values used for colormapping
lc.set_array(phi)
lc.set_linewidth(2)
line = axs[0].add_collection(lc)
axs[0].set_ylim((-1, 1))
axs[0].set_xlim((-1, 1))
axs[0].axis('equal')
fig.colorbar(line, ax=axs[0], label=r"$\varphi$ [rad]")
axs[0].grid(True)
axs[0].set_title("Matplotlib twilight")
axs[0].set_xlabel(r'$\Re\{e^{i\varphi}\}$')
axs[0].set_ylabel(r'$\Im\{e^{i\varphi}\}$')


lc = LineCollection(segments, cmap=spharpy.plot.phase_twilight(), norm=norm)
# # Set the values used for colormapping
lc.set_array(phi)
lc.set_linewidth(2)
line = axs[1].add_collection(lc)
axs[1].set_ylim((-1, 1))
axs[1].set_xlim((-1, 1))
axs[1].axis('equal')
fig.colorbar(line, ax=axs[1], label=r"$\varphi$ [rad]")
axs[1].grid(True)
axs[1].set_title("Spharpy twilight_phase")
axs[1].set_xlabel(r'$\Re\{e^{i\varphi}\}$')
axs[1].set_ylabel(r'$\Im\{e^{i\varphi}\}$')



import numpy as np
import numpy.testing as npt

import spharpy

def test_dolph_cheby_mainlobe():
    N = 7
    theta0 = np.pi/6
    d_nm = spharpy.beamforming.dolph_chebyshev_weights(
        N, theta0, design_criterion='mainlobe')

    truth = np.loadtxt('tests/data/dolph_cheby_mainlobe.csv', delimiter=',')
    npt.assert_allclose(d_nm, truth)


def test_dolph_cheby_sidelobe():
    N = 7
    R_dB = 50
    R = 10**(R_dB/20)
    d_nm = spharpy.beamforming.dolph_chebyshev_weights(
        N, R, design_criterion='sidelobe')

    truth = np.loadtxt('tests/data/dolph_cheby_sidelobe.csv', delimiter=',')
    npt.assert_allclose(d_nm, truth)


def test_re_max():
    N = 7
    g_nm = spharpy.beamforming.rE_max_weights(N)

    truth = np.loadtxt('tests/data/re_max_weights.csv', delimiter=',')
    npt.assert_allclose(g_nm, truth)


def test_max_front_back():
    N = 7
    f_nm = spharpy.beamforming.maximum_front_back_ratio_weights(N)

    truth = np.loadtxt('tests/data/max_front_back_weights.csv', delimiter=',')
    npt.assert_allclose(f_nm, truth)



import numpy as np
from spharpy.samplings.helpers import sph2cart
from scipy.spatial import cKDTree


class Coordinates(object):
    """Container class for coordinates in a three-dimensional space, allowing
    for compact representation and convenient conversion into spherical as well
    as geospatial coordinate systems.
    The constructor as well as the internal representation are only
    available in Cartesian coordinates. To create a Coordinates object from
    a set of points in spherical coordinates, please use the
    Coordinates.from_spherical() method.

    """
    def __init__(self, x=None, y=None, z=None):
        """Init coordinates container

        Parameters
        ----------
        x : ndarray, double
            x-coordinate
        y : ndarray, double
            y-coordinate
        z : ndarray, double
            z-coordinate
        """

        super(Coordinates, self).__init__()
        x = np.asarray(x, dtype=float)
        y = np.asarray(y, dtype=float)
        z = np.asarray(z, dtype=float)

        if not np.shape(x) == np.shape(y) == np.shape(z):
            raise ValueError("Input arrays need to have same dimensions.")

        self._x = x
        self._y = y
        self._z = z

    @property
    def x(self):
        """The x-axis coordinates for each point.
        """
        return self._x

    @x.setter
    def x(self, value):
        self._x = np.asarray(value, dtype=float)

    @property
    def y(self):
        """The y-axis coordinate for each point."""
        return self._y

    @y.setter
    def y(self, value):
        self._y = np.asarray(value, dtype=float)

    @property
    def z(self):
        """The z-axis coordinate for each point."""
        return self._z

    @z.setter
    def z(self, value):
        self._z = np.asarray(value, dtype=float)

    @property
    def radius(self):
        """The radius for each point."""
        return np.sqrt(self.x**2 + self.y**2 + self.z**2)

    @radius.setter
    def radius(self, radius):
        x, y, z = sph2cart(np.asarray(radius, dtype=float),
                           self.elevation,
                           self.azimuth)
        self._x = x
        self._y = y
        self._z = z

    @property
    def azimuth(self):
        """The azimuth angle for each point."""
        return np.mod(np.arctan2(self.y, self.x), 2*np.pi)

    @azimuth.setter
    def azimuth(self, azimuth):
        x, y, z = sph2cart(self.radius,
                           self.elevation,
                           np.asarray(azimuth, dtype=float))
        self._x = x
        self._y = y
        self._z = z

    @property
    def elevation(self):
        """The elevation angle for each point"""
        rad = self.radius
        return np.arccos(self.z/rad)

    @elevation.setter
    def elevation(self, elevation):
        x, y, z = sph2cart(self.radius,
                           np.asarray(elevation, dtype=float),
                           self.azimuth)
        self._x = x
        self._y = y
        self._z = z

    @classmethod
    def from_cartesian(cls, x, y, z):
        """Create a Coordinates class object from a set of points in the
        Cartesian coordinate system.

        Parameters
        ----------
        x : ndarray, double
            x-coordinate
        y : ndarray, double
            y-coordinate
        z : ndarray, double
            z-coordinate
        """
        return Coordinates(x, y, z)

    @classmethod
    def from_spherical(cls, radius, elevation, azimuth):
        """Create a Coordinates class object from a set of points in the
        spherical coordinate system.

        Parameters
        ----------
        radius : ndarray, double
            The radius for each point
        elevation : ndarray, double
            The elevation angle in radians
        azimuth : ndarray, double
            The azimuth angle in radians
        """
        radius = np.asarray(radius, dtype=float)
        elevation = np.asarray(elevation, dtype=float)
        azimuth = np.asarray(azimuth, dtype=float)
        x, y, z = sph2cart(radius, elevation, azimuth)
        return Coordinates(x, y, z)

    @classmethod
    def from_array(cls, values, coordinate_system='cartesian'):
        """Create a Coordinates class object from a set of points given as
        numpy array

        Parameters
        ----------
        values : double, ndarray
            Array with shape Nx3 where N is the number of points.
        coordinate_system : string
            Coordinate convention of the given values.
            Can be Cartesian or spherical coordinates.
        """
        coords = Coordinates()
        if coordinate_system == 'cartesian':
            coords.cartesian = values
        elif coordinate_system == 'spherical':
            coords.spherical = values
        else:
            return ValueError("This coordinate system is not supported.")

        return coords

    @property
    def latitude(self):
        """The latitude angle as used in geospatial coordinates."""
        return np.pi/2 - self.elevation

    @property
    def longitude(self):
        """The longitude angle as used in geospatial coordinates."""
        return np.arctan2(self.y, self.x)

    @property
    def cartesian(self):
        """Cartesian coordinates of all points."""
        return np.vstack((self.x, self.y, self.z))

    @cartesian.setter
    def cartesian(self, value):
        """Cartesian coordinates of all points."""
        self.x = value[0, :]
        self.y = value[1, :]
        self.z = value[2, :]

    @property
    def spherical(self):
        """Spherical coordinates of all points."""
        return np.vstack((self.radius, self.elevation, self.azimuth))

    @spherical.setter
    def spherical(self, value):
        """Cartesian coordinates of all points."""
        x, y, z = sph2cart(value[0, :], value[1, :], value[2, :])
        self.cartesian = np.vstack((x, y, z))

    @property
    def n_points(self):
        """Return number of points stored in the object"""
        return self.x.size

    def merge(self, other):
        """Merge another coordinates objects into this object."""
        data = np.concatenate(
            (self.cartesian, other.cartesian),
            axis=-1
        )
        self.cartesian = data

    def find_nearest_point(self, point):
        """Find the closest Coordinate point to a given Point.
        The search for the nearest point is performed using the scipy
        cKDTree implementation.

        Parameters
        ----------
        point : Coordinates
            Point to find nearest neighboring Coordinate

        Returns
        -------
        distance : ndarray, double
            Distance between the point and it's closest neighbor
        index : int
            Index of the closest point.

        """
        kdtree = cKDTree(self.cartesian.T)
        distance, index = kdtree.query(point.cartesian.T)

        return distance, index

    def __repr__(self):
        """repr for Coordinate class

        """
        if self.n_points == 1:
            repr_string = "Coordinates of 1 point"
        else:
            repr_string = "Coordinates of {} points".format(self.n_points)
        return repr_string

    def __getitem__(self, index):
        """Return Coordinates at index
        """
        return Coordinates(self._x[index], self._y[index], self._z[index])

    def __setitem__(self, index, item):
        """Set Coordinates at index
        """
        self.x[index] = item.x
        self.y[index] = item.y
        self.z[index] = item.z

    def __len__(self):
        """Length of the object which is the number of points stored.
        """
        return self.n_points


class SamplingSphere(Coordinates):
    """Class for samplings on a sphere"""

    def __init__(self, x=None, y=None, z=None, n_max=None, weights=None):
        """Init for sampling class
        """
        Coordinates.__init__(self, x, y, z)
        self._n_max = int(n_max) if n_max is not None else None
        if weights is None:
            self._weights = None

        elif len(x) != len(weights):
            raise ValueError("The number of weights has to be equal to \
                        the number of sampling points.")
        else:
            self._weights = np.asarray(weights, dtype=float)

    @property
    def n_max(self):
        """Spherical harmonic order."""
        return self._n_max

    @n_max.setter
    def n_max(self, value):
        self._n_max = int(value)

    @property
    def weights(self):
        """Sampling weights for numeric integration."""
        return self._weights

    @weights.setter
    def weights(self, weights):
        if len(weights) != self.n_points:
            raise ValueError("The number of weights has to be equal to \
                    the number of sampling points.")
        self._weights = np.asarray(weights, dtype=float)

    @classmethod
    def from_coordinates(cls, coords, n_max=None, weights=None):
        """Generate a spherical sampling object from a coordinates object

        Parameters
        ----------
        coords : Coordinates
            Coordinate object

        Returns
        -------
        sampling : SamplingSphere
            Sampling on a sphere

        """
        return SamplingSphere(coords.x, coords.y, coords.z,
                              n_max=n_max, weights=weights)

    @classmethod
    def from_cartesian(cls, x, y, z, n_max=None, weights=None):
        """Create a Coordinates class object from a set of points in the
        Cartesian coordinate system.

        Parameters
        ----------
        x : ndarray, double
            x-coordinate
        y : ndarray, double
            y-coordinate
        z : ndarray, double
            z-coordinate
        """
        return SamplingSphere(x, y, z, n_max, weights)

    @classmethod
    def from_spherical(
            cls, radius, elevation, azimuth,
            n_max=None, weights=None):
        """Create a Coordinates class object from a set of points in the
        spherical coordinate system.

        Parameters
        ----------
        radius : ndarray, double
            The radius for each point
        elevation : ndarray, double
            The elevation angle in radians
        azimuth : ndarray, double
            The azimuth angle in radians
        """
        radius = np.asarray(radius, dtype=float)
        elevation = np.asarray(elevation, dtype=float)
        azimuth = np.asarray(azimuth, dtype=float)
        x, y, z = sph2cart(radius, elevation, azimuth)
        return SamplingSphere(x, y, z, n_max, weights)

    @classmethod
    def from_array(
            cls, values, n_max=None, weights=None,
            coordinate_system='cartesian'):
        """Create a Coordinates class object from a set of points given as
        numpy array

        Parameters
        ----------
        values : double, ndarray
            Array with shape Nx3 where N is the number of points.
        coordinate_system : string
            Coordinate convention of the given values.
            Can be Cartesian or spherical coordinates.
        """
        coords = SamplingSphere(n_max=n_max, weights=weights)
        if coordinate_system == 'cartesian':
            coords.cartesian = values
        elif coordinate_system == 'spherical':
            coords.spherical = values
        else:
            return ValueError("This coordinate system is not supported.")

        return coords

    def __repr__(self):
        """repr for SamplingSphere class
        """
        if self.n_points == 1:
            repr_string = "Sampling with {} point".format(self.n_points)
        else:
            repr_string = "Sampling with {} points".format(self.n_points)
        return repr_string



import numpy as np
import scipy.spatial as sspat


def greens_function_plane_wave(
        source_points,
        receiver_points,
        wave_number,
        gradient=False):
    """The matrix describing the propagation of a plane wave from a direction
    of arrival defined by the azimuth and elevation angles of the source points
    to the receiver points. The phase sign convention reflects a direction of
    arrival from the source position.

    Parameters
    ----------
    source_points : Coordinates
        The source points defining the direction of incidence for the plane
        wave. Note that the radius on which the source is positioned has no
        relevance.
    receiver_points : Coordinates
        The receiver points.
    wave_number : double
        The wave number of the wave
    speed_of_sound : double
        The speed of sound
    gradient : bool
        If True, the gradient will be returned as well


    Returns
    -------
    M : ndarray, complex, shape(n_receiver, n_sources)
        The plane wave propagation matrix

    """
    e_doa = source_points.cartesian / \
        np.linalg.norm(source_points.cartesian, axis=0)
    k_vec = np.squeeze(wave_number*e_doa)

    # avoid using the complex exponential since it's slower than sin and cos
    arg = receiver_points.cartesian.T @ k_vec
    plane_wave_matrix = np.cos(arg) + 1j*np.sin(arg)

    if not gradient:
        return plane_wave_matrix
    else:
        plane_wave_gradient_matrix_x = (plane_wave_matrix * 1j*k_vec[0])
        plane_wave_gradient_matrix_y = (plane_wave_matrix * 1j*k_vec[1])
        plane_wave_gradient_matrix_z = (plane_wave_matrix * 1j*k_vec[2])
        return plane_wave_matrix, \
            [plane_wave_gradient_matrix_x,
             plane_wave_gradient_matrix_y,
             plane_wave_gradient_matrix_z]


def greens_function_point_source(sources, receivers, k, gradient=False):
    r"""Green's function for point sources in free space in matrix form. The
    phase sign convention corresponds to a direction of propagation away from
    the source at position $r_s$.

    .. math::

        G(k) = \\frac{e^{- k\\|\\mathbf{r_s} - \\mathbf{r_r}\\|}}
            {4 \\pi \\|\\mathbf{r_s} - \\mathbf{r_r}\\|}

    Parameters
    ----------
    source : Coordinates
        source points as Coordinates object
    receivers : Coordinates
        receiver points as Coordinates object
    k : ndarray, double
        wave number

    Returns
    -------
    G : ndarray, double
        Green's function

    """
    dist = sspat.distance.cdist(receivers.cartesian.T, sources.cartesian.T)
    dist = np.squeeze(dist)
    cexp = np.cos(k*dist) - 1j*np.sin(k*dist)
    G = cexp/dist/4/np.pi

    if not gradient:
        return G
    else:
        def lambda_cdiff(u, v):
            return u-v

        diff_x = sspat.distance.cdist(
            np.atleast_2d(receivers.x).T,
            np.atleast_2d(sources.x).T,
            lambda_cdiff)
        diff_y = sspat.distance.cdist(
            np.atleast_2d(receivers.y).T,
            np.atleast_2d(sources.y).T,
            lambda_cdiff)
        diff_z = sspat.distance.cdist(
            np.atleast_2d(receivers.z).T,
            np.atleast_2d(sources.z).T,
            lambda_cdiff)

        G_dx = G/dist * np.squeeze(diff_x) * (-1j-1/dist)
        G_dy = G/dist * np.squeeze(diff_y) * (-1j-1/dist)
        G_dz = G/dist * np.squeeze(diff_z) * (-1j-1/dist)

        return G, (G_dx, G_dy, G_dz)



from .rotations import (
    rotation_z_axis,
    rotation_z_axis_real,
    wigner_d_function,
    wigner_d_rotation,
    wigner_d_rotation_real,
    RotationSH)


__all__ = [
    'rotation_z_axis',
    'rotation_z_axis_real',
    'wigner_d_rotation',
    'wigner_d_rotation_real',
    'wigner_d_function',
    'RotationSH']



from matplotlib import cm
from matplotlib.colors import ListedColormap
import numpy as np


def phase_twilight(lut=512):
    """Cyclic color map for displaying phase information. This is a modified
    version of the twilight color map from matplotlib.
    """
    lut = int(np.ceil(lut/4)*4)
    twilight = cm.get_cmap('twilight', lut=lut)

    twilight_r_colors = np.array(twilight.reversed().colors)

    roll_by = int(lut/4)
    phase_colors = np.roll(twilight_r_colors, -roll_by, axis=0)

    return ListedColormap(phase_colors)



"""
Rotation/Translation operations for data in the spherical harmonic domains
"""

import numpy as np
import spharpy
from scipy.special import eval_jacobi, factorial
from scipy.spatial.transform import Rotation


class RotationSH(Rotation):
    """Class for rotations of coordinates and data.

    """

    def __init__(self, quat, n_max=0, *args, **kwargs):
        """Initialize

        Parameters
        ----------
        quat : array_like, shape (N, 4) or (4,)
            Each row is a (possibly non-unit norm) quaternion in scalar-last
            (x, y, z, w) format. Each quaternion will be normalized to unit
            norm.
        n_max : int
            The spherical harmonic order

        Returns
        -------
        RotationSH
            The rotation object with spherical harmonic order n_max.

        Note
        ----
        Initializing using the constructor is not advised. Always use the
        respective ``from_quat`` method.

        """
        super().__init__(quat, *args, **kwargs)
        if n_max < 0:
            raise ValueError("The order needs to be a positive value.")
        self._n_max = int(n_max)

    @classmethod
    def from_rotvec(cls, n_max, rotvec, degrees=False, *args, **kwargs):
        """Initialize from rotation vectors.
        A rotation vector is a 3 dimensional vector which is co-directional to
        the axis of rotation and whose norm gives the angle of rotation [#]_.

        Parameters
        ----------
        n_max : int
            Spherical harmonic order
        rotvec : array_like, float, shape (L, 3) or (3,)
            A single vector or a stack of vectors, where rot_vec[i] gives the
            ith rotation vector.
        degrees : bool, optional
            Specify if rotation angles are defined in degrees instead of
            radians, by default False.

        Returns
        -------
        RotationSH
            Object containing the rotations represented by input rotation
            vectors.

        References
        ----------
        .. [#] https://en.wikipedia.org/wiki/Axis%E2%80%93angle_representation\
#Rotation_vector

        Examples
        --------
        >>> from spharpy.transforms import Rotation as R
        >>> rot = R.from_rotvec(np.pi/2 * np.array([0, 0, 1]))
        >>> rot.as_spherical_harmonic()

        """
        if degrees:
            rotvec = np.deg2rad(rotvec)

        cls = super(RotationSH, cls).from_rotvec(rotvec, *args, **kwargs)
        cls.n_max = n_max
        return cls

    @classmethod
    def from_euler(cls, n_max, seq, angles, degrees=False, **kwargs):
        """Initialize from Euler angles.

        Rotations in 3-D can be represented by a sequence of 3 rotations
        around a sequence of axes. In theory, any three axes spanning the 3-D
        Euclidean space are enough. In practice, the axes of rotation are
        chosen to be the basis vectors.
        The three rotations can either be in a global frame of reference
        (extrinsic) or in a body centred frame of reference (intrinsic), which
        is attached to, and moves with, the object under rotation [#]_.

        Parameters
        ----------
        n_max : int
            Spherical harmonic order.
        seq : str
            Specifies sequence of axes for rotations. Up to 3 characters
            belonging to the set {‘X’, ‘Y’, ‘Z’} for intrinsic rotations,
            or {‘x’, ‘y’, ‘z’} for extrinsic rotations. Extrinsic and intrinsic
            rotations cannot be mixed in one function call.

        angles : (float or array_like, shape (N,) or (N, [1 or 2 or 3]))
            Euler angles specified in radians (degrees is False) or degrees
            (degrees is True). For a single character seq, angles can be:

            - a single value
            - array_like with shape (N,), where each ``angle[i]`` corresponds
              to a single rotation
            - array_like with shape (N, 1), where each ``angle[i, 0]``
              corresponds to a single rotation

            For 2- and 3-character wide seq, angles can be:

            - array_like with shape (W,) where W is the width of ``seq``, which
              corresponds to a single rotation with W axes
            - array_like with shape (N, W) where each ``angle[i]`` corresponds
              to a sequence of Euler angles describing a single rotation

        degrees : bool, optional
            If True, then the given angles are assumed to be in degrees.
            Default is False.

        Returns
        -------
        RotationSH
            Object containing the rotation represented by the sequence of
            rotations around given axes with given angles.

        References
        ----------
        .. [#] https://en.wikipedia.org/wiki/Euler_angles

        Examples
        --------
        >>> from spharpy.transforms import Rotation as R
        >>> rot = R.from_euler('z', 90, degrees=True)
        >>> rot.as_spherical_harmonic()


        """
        cls = super(RotationSH, cls).from_euler(
            seq, angles, degrees=degrees, **kwargs)
        cls.n_max = n_max
        return cls

    @classmethod
    def from_quat(cls, n_max, quat, **kwargs):
        """Initialize from quaternions.
        3D rotations can be represented using unit-norm quaternions [#]_.

        Parameters
        ----------
        n_max : int
            Spherical harmonic order
        quat : (array_like, shape (N, 4) or (4,))
            Each row is a (possibly non-unit norm) quaternion in scalar-last
            (x, y, z, w) format. Each quaternion will be normalized to unit
            norm.

        Returns
        -------
        RotationSH
            Object containing the rotations represented by input quaternions.

        References
        ----------
        .. [#] https://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation

        Examples
        --------
        >>> from spharpy.transforms import Rotation as R
        >>> rot = R.from_quat([1, 0, 0, 0])
        >>> rot.as_spherical_harmonic()

        """
        cls = super(RotationSH, cls).from_quat(quat, **kwargs)
        cls.n_max = n_max
        return cls

    @classmethod
    def from_matrix(cls, n_max, matrix, **kwargs):
        """Initialize from rotation matrix.
        Rotations in 3 dimensions can be represented with 3 x 3 proper
        orthogonal matrices [1]_. If the input is not proper orthogonal,
        an approximation is created using the method described in [2]_.

        Parameters
        ----------
        n_max : int
            Spherical harmonic order
        matrix : (array_like, shape (N, 3, 3) or (3, 3))
            A single matrix or a stack of matrices, where ``matrix[i]`` is
            the i-th matrix.

        Returns
        -------
        RotationSH
            Object containing the rotations represented by the rotation
            matrices.

        References
        ----------
        .. [1] https://en.wikipedia.org/wiki/Rotation_matrix
        .. [2] F. Landis Markley, “Unit Quaternion from Rotation Matrix”,
               Journal of guidance, control, and dynamics vol. 31.2,
               pp. 440-442, 2008.

        Examples
        --------
        >>> from spharpy.transforms import Rotation as R
        >>> rot = R.from_matrix([
        ...     [0, -1, 0],
        ...     [1,  0, 0],
        ...     [0,  0, 1]])
        >>> rot.as_spherical_harmonic()
        """
        cls = super(RotationSH, cls).from_matrix(matrix, **kwargs)
        cls.n_max = n_max
        return cls

    @property
    def n_max(self):
        """The spherical harmonic order used for spherical harmonic rotation
        matrices.
        """
        return self._n_max

    @n_max.setter
    def n_max(self, value):
        """Set the spherical harmonic order

        Parameters
        ----------
        value : int
            The spherical harmonic order used for spherical harmonic rotation
        matrices.
        """
        if value < 0:
            raise ValueError("The order needs to be a positive value.")
        self._n_max = value

    def as_spherical_harmonic(self, type='real'):
        """Export the rotation operations as a spherical harmonic rotation
        matrices. Supports complex and real-valued spherical harmonics.

        Parameters
        ----------
        real : string, optional
            Spherical harmonic definition. Can either be 'complex' or 'real',
            by default 'real' is used.

        Returns
        -------
        array, complex or float
            Stack of block-diagonal rotation matrices.
        """
        euler_angles = np.atleast_2d(self.as_euler('zyz'))
        n_matrices = euler_angles.shape[0]

        n_sh = (self.n_max+1)**2
        if type == 'real':
            dtype = float
            rot_func = wigner_d_rotation_real
        elif type == 'complex':
            dtype = complex
            rot_func = wigner_d_rotation
        else:
            raise ValueError("Invalid spherical harmonic type {}".format(type))

        D = np.zeros((n_matrices, n_sh, n_sh), dtype=dtype)

        for idx, angles in enumerate(euler_angles):
            D[idx, :, :] = rot_func(
                self.n_max, angles[0], angles[1], angles[2])

        return np.squeeze(D)

    def apply(self, coefficients, type='real'):
        """Apply the rotation to L sets of spherical harmonic coefficients

        Parameters
        ----------
        coefficients : array, complex, shape :math:`((n_max+1)^2, L)`
            L sets of spherical harmonic coefficients with a respective order
            :math:`((n_max+1)^2`

        Returns
        -------
        array, complex
            The rotated data
        """
        D = self.as_spherical_harmonic(type=type)
        if D.ndim > 2:
            M = np.diag(np.ones((self.n_max+1)**2))
            for d in D:
                M = M @ d
        else:
            M = D

        return M @ coefficients


def rotation_z_axis(n_max, angle):
    """Rotation matrix for complex spherical harmonics around the z-axis
    by a given angle. The rotation is performed such that positive angles
    result in a counter clockwise rotation of the data [#]_.

    .. math::

        c_{nm}(\\theta, \\phi + \\xi) = e^{-im\\xi} c_{nm}(\\theta, \\phi)

    Parameters
    ----------
    n_max : integer
        Spherical harmonic order
    angle : number
        Rotation angle in radians `[0, 2 \\pi]`

    Returns
    -------
    array_like, complex, shape :math:`((n_{max}+1)^2, (n_{max}+1)^2)`
        Diagonal rotation matrix evaluated for the specified angle

    References
    ----------
    .. [#]  N. A. Gumerov and R. Duraiswami, “Recursions for the computation
            of multipole translation and rotation coefficients for the 3-d
            helmholtz equation,” vol. 25, no. 4, pp. 1344–1381, 2003.


    Examples
    --------
    >>> import numpy as np
    >>> import spharpy
    >>> n_max = 1
    >>> sh_vec = np.array([0, 1, 0, 0])
    >>> rotMat = spharpy.transforms.rotation_z_axis(n_max, np.pi/2)
    >>> sh_vec_rotated = rotMat @ sh_vec
    """

    acn = np.arange(0, (n_max+1)**2)
    m = spharpy.spherical.acn2nm(acn)[1]
    rotation_phi = np.exp(-1j*angle*m)

    return np.diag(rotation_phi)


def rotation_z_axis_real(n_max, angle):
    """Rotation matrix for real-valued spherical harmonics around the z-axis
    by a given angle. The rotation is performed such that positive angles
    result in a counter clockwise rotation of the data [#]_.

    Parameters
    ----------
    n_max : integer
        Spherical harmonic order
    angle : number
        Rotation angle in radians `[0, 2 \\pi]`

    Returns
    -------
    array_like, float, shape :math:`((n_max+1)^2, (n_max+1)^2)`
        Block-diagonal Rotation matrix evaluated for the specified angle.

    References
    ----------
    .. [#] M. Kronlachner, “Spatial Transformations for the Alteration of
           Ambisonic Recordings,” Master Thesis, University of Music and
           Performing Arts, Graz, 2014.


    Examples
    --------
    >>> import numpy as np
    >>> import spharpy
    >>> n_max = 1
    >>> sh_vec = np.array([0, 1, 0, 0])
    >>> rotMat = spharpy.transforms.rotation_z_axis_real(n_max, np.pi/2)
    >>> sh_vec_rotated = rotMat @ sh_vec

    """
    acn = np.arange(0, (n_max + 1) ** 2)
    n, m = spharpy.spherical.acn2nm(acn)
    acn_reverse_degree = n ** 2 + n - m

    rotation_phi = np.zeros(((n_max + 1) ** 2, (n_max + 1) ** 2))
    mask = m == 0
    rotation_phi[acn[mask], acn[mask]] = 1.0

    mask_pos = m > 0
    mask_neg = m < 0
    rotation_phi[acn[mask_pos], acn[mask_pos]] = np.cos(
        np.abs(m[mask_pos]) * angle)
    rotation_phi[acn[mask_neg], acn[mask_neg]] = np.cos(
        np.abs(m[mask_neg]) * angle)

    # non diagonal
    rotation_phi[acn[mask_pos], acn_reverse_degree[mask_pos]] = -np.sin(
        np.abs(m[mask_pos]) * angle)
    rotation_phi[acn[mask_neg], acn_reverse_degree[mask_neg]] = np.sin(
        np.abs(m[mask_neg]) * angle)

    return rotation_phi


def wigner_d_rotation(n_max, alpha, beta, gamma):
    r"""Wigner-D rotation matrix for Euler rotations by angles
    (\alpha, \beta, \gamma) around the (z,y,z)-axes.
    The implementation follows [#]_. and rotation is performed such that
    positive angles result in a counter clockwise rotation of the data.

    .. math::

        D_{m^\prime,m}^n(\alpha, \beta, \gamma) =
        e^{-im^\prime\alpha} d_{m^\prime,m}^n(\beta) e^{-im\gamma}

    Parameters
    ----------
    n_max : int
        Spherical harmonic order
    alpha : float
        First z-axis rotation angle
    beta : float
        Y-axis rotation angle
    gamma : float
        Second z-axis rotation angle

    Returns
    -------
    array_like, complex,:math:`((n_max+1)^2, (n_max+1)^2)`
        Block diagonal rotation matrix


    Examples
    --------
    >>> import numpy as np
    >>> import spharpy
    >>> n_max = 1
    >>> sh_vec = np.array([0, 0, 1, 0])
    >>> rotMat = spharpy.transforms.wigner_d_rotation(n_max, 0, np.pi/4, 0)
    >>> sh_vec_rotated = rotMat @ sh_vec

    References
    ----------
    .. [#] B. Rafaely, Fundamentals of Spherical Array Processing, 1st ed.,
           vol. 8. Springer-Verlag GmbH Berlin Heidelberg, 2015.

    """

    n_sh = (n_max+1)**2

    R = np.zeros((n_sh, n_sh), dtype=complex)

    for row in np.arange(0, (n_max+1)**2):
        n_dash, m_dash = spharpy.spherical.acn2nm(row)
        for column in np.arange(0, (n_max+1)**2):
            n, m = spharpy.spherical.acn2nm(column)
            if n == n_dash:
                rot_alpha = np.exp(-1j*m_dash*alpha)
                rot_beta = wigner_d_function(n, m_dash, m, beta)
                rot_gamma = np.exp(-1j*m*gamma)
                R[row, column] = rot_alpha * rot_beta * rot_gamma

    return R


def wigner_d_rotation_real(n_max, alpha, beta, gamma):
    r"""Wigner-D rotation matrix for Euler rotations for real-valued spherical
    harmonics by angles (\alpha, \beta, \gamma) around the (z,y,z)-axes.
    The implementation follows [#]_ and the rotation is performed such that
    positive angles result in a counter clockwise rotation of the data.

    Parameters
    ----------
    n_max : int
        Spherical harmonic order
    alpha : float
        First z-axis rotation angle
    beta : float
        Y-axis rotation angle
    gamma : float
        Second z-axis rotation angle

    Returns
    -------
    array_like, float, :math:`((n_max+1)^2, (n_max+1)^2)`
        Block diagonal rotation matrix

    Examples
    --------
    >>> import numpy as np
    >>> import spharpy
    >>> n_max = 1
    >>> sh_vec = np.array([0, 0, 1, 0])
    >>> rotMat = spharpy.transforms.wigner_d_rotation_real(
    >>>     n_max, 0, np.pi/4, 0)
    >>> sh_vec_rotated = rotMat @ sh_vec

    References
    ----------
    .. [#]  M. A. Blanco, M. Flórez, and M. Bermejo, “Evaluation of the
            rotation matrices in the basis of real spherical harmonics,”
            Journal of Molecular Structure: THEOCHEM,  vol. 419, no. 1–3,
            pp. 19–27, Dec. 1997, doi: 10.1016/S0166-1280(97)00185-1.

    """
    n_sh = (n_max+1)**2
    R = np.zeros((n_sh, n_sh), dtype=float)

    for row_acn in np.arange(0, (n_max+1)**2):
        for col_acn in np.arange(0, (n_max+1)**2):
            n, m = spharpy.spherical.acn2nm(col_acn)
            n_dash, m_dash = spharpy.spherical.acn2nm(row_acn)
            if n == n_dash:
                # minus beta opposite rotation direction
                d_l_1 = wigner_d_function(n, np.abs(m_dash), np.abs(m), -beta)
                d_l_2 = wigner_d_function(n, np.abs(m), -np.abs(m_dash), -beta)

                R[row_acn, col_acn] = \
                    _sign(m_dash) * _Phi(m, alpha) * _Phi(m_dash, gamma) * \
                    (d_l_1 + (-1)**int(m) * d_l_2)/2 \
                    - _sign(m) * _Phi(-m, alpha) * _Phi(-m_dash, gamma) * \
                    (d_l_1 - (-1)**int(m) * d_l_2)/2

    return R


def _sign(x):
    """
    Returns sign of x, differs from numpy definition for x=0
    """
    if x < 0:
        sign = -1
    else:
        sign = 1

    return sign


def _Phi(m, angle):
    """
    Rotation Matrix around z-axis for real Spherical Harmonics as defined in
    Blanco et al., Evaluation of the rotation matrices in the basis of real
    spherical harmonics, eq.(8)
    """
    if m > 0:
        phi = np.sqrt(2)*np.cos(m*angle)
    elif m == 0:
        phi = 1
    elif m < 0:
        # minus due to differing phase convention
        phi = -np.sqrt(2)*np.sin(np.abs(m)*angle)

    return phi


def wigner_d_function(n, m_dash, m, beta):
    r"""Wigner-d function for rotations around the y-axis.
    Convention as defined in [#]_.

    Parameters
    ----------
    n : int
        order
    m_dash : int
        degree
    m : int
        degree
    beta : float
        Rotation angle

    Returns
    -------
    float
        Wigner-d symbol

    References
    ----------
    .. [#] B. Rafaely, Fundamentals of Spherical Array Processing, 1st ed.,
           vol. 8. Springer-Verlag GmbH Berlin Heidelberg, 2015.

    """

    if m >= m_dash:
        sign = 1
    else:
        sign = (-1)**int(m-m_dash)

    mu = np.abs(m_dash - m)
    nu = np.abs(m_dash + m)
    s = n - (mu + nu)/2

    norm = (factorial(s)*factorial(s+mu+nu))/(factorial(s+mu)*factorial(s+nu))
    P = eval_jacobi(s, mu, nu, np.cos(beta))
    d = sign * np.sqrt(norm) * np.sin(beta/2)**mu * np.cos(beta/2)**nu * P

    return d



import numpy as np
import scipy.special as spspecial
from spharpy.special import spherical_bessel_zeros
from spharpy.samplings import Coordinates


def interior_stabilization_points(kr_max, resolution_factor=1):
    """ Find points inside the interior domain of an open  spherical microphone
    array that stabilize the array array response at the eigenfrequencies of
    the array. The algorithm is based on [7]_ and implemented following the
    Matlab code provided by Gilles Chardon on his homepage at [8]_.
    The stabilization points are independent of the sampling of the sphere
    and can therefore be combined with arbitrary spherical samplings.

    Parameters
    ----------
    kr_max : float
        The maximum kr value to be considered. This will define
        the upper frequency limit of the array.
    resolution_factor : int
        Factor to increase the spatial resolution of the grid
        used to estimate the stabilization points.

    Returns
    -------
    sampling_interior : Coordinates
        Coordinates of the stabilization points

    References
    ----------
    .. [7]  G. Chardon, W. Kreuzer, und M. Noisternig, "Design of spatial
            microphone arrays for sound field interpolation", IEEE Journal of
            Selected Topics in Signal Processing
    .. [8]  https://gilleschardon.fr/jstsp_array/

    """
    x, y, z = find_interior_points(kr_max, resolution_factor=resolution_factor)

    return Coordinates(x, y, z)


def find_eigenfrequencies(kr_max):
    """
    Find the eigenfrequencies for the sphere from the spherical
    Bessel function.
    """
    jn_zeros = spherical_bessel_zeros(kr_max, kr_max)

    eigenfrequencies = []
    for idx in range(kr_max + 1):
        roots = jn_zeros[idx]
        roots = roots[roots < kr_max]
        if roots.size != 0:
            eigenfrequencies.append(roots)

    mults = np.arange(1, 2*len(eigenfrequencies))

    return eigenfrequencies, mults


def calculate_eigenspaces(kr_max, theta, phi, rad):
    """Calculate the eigenspaces for the corresponding eigenfrequencies of
    the sphere

    Parameters
    ----------
    k_max : float
        The largest wave number to be included
    theta : array, float
        Azimuth angle
    phi : array, float
        Elevation angle
    rad : array, float
        Radius

    Returns
    -------
    eigenspaces : list, ndarray, float
        List containing all eigenspaces

    """
    eigenfrequencies, mults = find_eigenfrequencies(kr_max)

    subspaces = []
    for u in range(len(eigenfrequencies)):
        subspaces.extend(
            sph_modes_matrix(u, root, theta, phi, rad)
            for root in eigenfrequencies[u])
    return subspaces, mults


def sph_modes_matrix(n_max, k, theta, phi, rad):
    """Build the matrix containing all spherical harmonic modes of the domain
    inside an open sphere for a specific order n_max.

    Parameters
    ----------
    n_max : int
        Spherical harmonic order
    k : float
        Wave number
    theta : array, float
        Azimuth angle
    phi : array, float
        Elevation angle
    rad : array, float
        Radius

    Returns
    -------
    modes : array, float
        A matrix with dimension [(...) x (2*n_max+1)] containing all spherical
        harmonic modes.

    Note
    ----
    This function returns only coefficients for one order, but all degrees.

    """
    n_coefficients = 2*n_max+1
    meshgrid_shape = theta.shape
    B = spspecial.spherical_jn(n_max, rad.flatten()*k) * 4*np.pi * (1j)**n_max
    B = np.reshape(B, meshgrid_shape)
    M = np.zeros((*meshgrid_shape, n_coefficients), dtype=complex)
    for m in range(-n_max, n_max+1):
        Y_m = spspecial.sph_harm(m, n_max, theta.flatten(), phi.flatten())
        M[:, :, :, m+n_max] = B * np.reshape(Y_m, meshgrid_shape)

    return M


def ball_dot(S1, S2, radius, phi):
    wphi = np.sin(phi)
    wr = radius**2
    w = wr*wphi
    return np.sum(np.conj(S1) * S2 * w) / np.sum(w)


def find_interior_points(k_max, resolution_factor=1):
    resolution = 50 * resolution_factor

    vec_theta = np.linspace(0, 2 * np.pi, resolution*2)
    vec_phi = np.linspace(0, np.pi, resolution)
    vec_rad = np.linspace(0, 1, resolution)

    phi, theta, rad = np.meshgrid(vec_phi, vec_theta, vec_rad)
    meshgrid_shape = (resolution*2, resolution, resolution)

    subspaces, mults = calculate_eigenspaces(k_max, theta, phi, rad)
    max_mult = mults.max()

    idx_sel = []
    for _ in range(max_mult):
        maxes = np.ones((*meshgrid_shape, len(subspaces))) * 1e3

        for idx_space in range(len(subspaces)):
            if subspaces[idx_space].size > 0:
                maxes[:, :, :, idx_space] = 0.0

                for v in range(subspaces[idx_space].shape[3]):
                    vector = subspaces[idx_space][:, :, :, v]
                    for www in range(v):
                        vector = vector - ball_dot(
                                        subspaces[idx_space][:, :, :, www],
                                        vector,
                                        rad,
                                        phi) * \
                                    subspaces[idx_space][:, :, :, www]
                    vector = vector / np.sqrt(np.abs(ball_dot(vector,
                                                              vector,
                                                              rad,
                                                              phi)))

                    subspaces[idx_space][:, :, :, v] = vector
                    maxes[:, :, :, idx_space] += np.abs(vector) ** 2

        minmax = np.min(maxes, axis=3)
        argmax = np.argmax(minmax)
        argmax_unravel = np.unravel_index(argmax, meshgrid_shape)
        idx_sel.append(argmax)

        for idx_space in range(len(subspaces)):
            if subspaces[idx_space].shape[3] <= 1:
                subspaces[idx_space] = np.array([[[[]]]])
            else:
                value_end = 0
                www = -1
                while value_end == 0:
                    www += 1
                    vend = subspaces[idx_space][:, :, :, www].copy()
                    value_end = vend[argmax_unravel]

                subspaces[idx_space][:, :, :, www] = \
                    subspaces[idx_space][:, :, :, -1]
                subspaces[idx_space][:, :, :, -1] = vend

                for v in range(subspaces[idx_space].shape[3] - 1):
                    vv = subspaces[idx_space][:, :, :, v]
                    value_vv = vv[argmax_unravel]
                    subspaces[idx_space][:, :, :, v] = \
                        vv - vend / value_end * value_vv

                subspaces[idx_space] = subspaces[idx_space][:, :, :, :-1]

    idx_sel_unravel = np.unravel_index(idx_sel, meshgrid_shape)
    theta_opt = theta[idx_sel_unravel]
    phi_opt = phi[idx_sel_unravel]
    rad_opt = rad[idx_sel_unravel]

    x = rad_opt * np.cos(theta_opt) * np.sin(phi_opt)
    y = rad_opt * np.sin(theta_opt) * np.sin(phi_opt)
    z = rad_opt * np.cos(phi_opt)

    return x, y, z



"""
Collection of sampling schemes for the sphere
"""
import urllib3
import numpy as np
from spharpy.samplings.coordinates import Coordinates, SamplingSphere
import spharpy

from ._eqsp import point_set as eq_point_set


def cube_equidistant(n_points):
    """Create a cuboid sampling with equidistant spacings in x, y, and z.
    The cube will have dimensions 1 x 1 x 1

    Parameters
    ----------
    n_points : int, tuple
        Number of points in the sampling. If a single value is given, the
        number of sampling positions will be the same in every axis. If a tuple
        is given, the number of points will be set as (n_x, n_y, n_z)

    Returns
    -------
    sampling : Coordinates
        Sampling positions as Coordinate object

    """
    if np.size(n_points) == 1:
        n_x = n_points
        n_y = n_points
        n_z = n_points
    elif np.size(n_points) == 3:
        n_x = n_points[0]
        n_y = n_points[1]
        n_z = n_points[2]
    else:
        raise ValueError("The number of points needs to be either an integer \
                or a tuple with 3 elements.")

    x = np.linspace(-1, 1, n_x)
    y = np.linspace(-1, 1, n_y)
    z = np.linspace(-1, 1, n_z)

    x_grid, y_grid, z_grid = np.meshgrid(x, y, z)

    return Coordinates(x_grid.flatten(), y_grid.flatten(), z_grid.flatten())


def hyperinterpolation(n_max):
    """Gives the points of a Hyperinterpolation sampling grid
    after Sloan and Womersley [1]_.

    Notes
    -----
    This implementation uses precalculated sets of points which are downloaded
    from Womersley's homepage [2]_.

    References
    ----------
    .. [1]  I. H. Sloan and R. S. Womersley, “Extremal Systems of Points and
            Numerical Integration on the Sphere,” Advances in Computational
            Mathematics, vol. 21, no. 1/2, pp. 107–125, 2004.
    .. [2]  http://web.maths.unsw.edu.au/~rsw/Sphere/Extremal/New/index.html

    Parameters
    ----------
    n_max : integer
        Spherical harmonic order of the sampling

    Returns
    -------
    sampling: SamplingSphere
        SamplingSphere object containing all sampling points
    """
    n_sh = (n_max+1)**2
    filename = "md%03d.%05d" % (n_max, n_sh)
    url = "https://web.maths.unsw.edu.au/~rsw/Sphere/S2Pts/MD/"
    fileurl = url + filename

    http = urllib3.PoolManager(cert_reqs=False)
    http_data = http.urlopen('GET', fileurl)

    if http_data.status == 200:
        file_data = http_data.data.decode()
    else:
        raise ConnectionError("Connection error. Please check your internet \
                connection.")

    file_data = np.fromstring(
        file_data, dtype='double', sep=' ').reshape((n_sh, 4))
    sampling = SamplingSphere(
        file_data[:, 0],
        file_data[:, 1],
        file_data[:, 2])
    sampling.weights = file_data[:, 3]

    return sampling


def spherical_t_design(n_max, criterion='const_energy'):
    r"""Return the sampling positions for a spherical t-design [3]_ .
    For a spherical harmonic order N, a t-Design of degree `:math: t=2N` for
    constant energy or `:math: t=2N+1` additionally ensuring a constant angular
    spread of energy is required [4]_. For a given degree t

    .. math::

        L = \lceil \frac{(t+1)^2}{2} \rceil+1,

    points will be generated, except for t = 3, 5, 7, 9, 11, 13, and 15.
    T-designs allow for a inverse spherical harmonic transform matrix
    calculated as `:math: D = \frac{4\pi}{L} \mathbf{Y}^\mathrm{H}`.

    Parameters
    ----------
    degree : integer
        T-design degree
    criterion : 'const_energy', 'const_angular_spread'
        Design criterion ensuring only a constant energy or additionally
        constant angular spread of energy

    Returns
    -------
    sampling : SamplingSphere
        SamplingSphere object containing all sampling points

    Notes
    -----
    This function downloads a pre-calculated set of points from
    Rob Womersley's homepage [5]_ .

    References
    ----------

    .. [3]  C. An, X. Chen, I. H. Sloan, and R. S. Womersley, “Well Conditioned
            Spherical Designs for Integration and Interpolation on the
            Two-Sphere,” SIAM Journal on Numerical Analysis, vol. 48, no. 6,
            pp. 2135–2157, Jan. 2010.
    .. [4]  F. Zotter, M. Frank, and A. Sontacchi, “The Virtual T-Design
            Ambisonics-Rig Using VBAP,” in Proceedings on the Congress on
            Sound and Vibration, 2010.
    .. [5]  http://web.maths.unsw.edu.au/~rsw/Sphere/EffSphDes/sf.html

    """
    if criterion == 'const_energy':
        degree = 2*n_max
    elif criterion == 'const_angular_spread':
        degree = 2*n_max + 1
    else:
        raise ValueError("Invalid design criterion.")

    n_points = int(np.ceil((degree + 1)**2 / 2) + 1)
    n_points_exceptions = {3: 8, 5: 18, 7: 32, 9: 50, 11: 72, 13: 98, 15: 128}
    if degree in n_points_exceptions:
        n_points = n_points_exceptions[degree]

    filename = "sf%03d.%05d" % (degree, n_points)
    url = "http://web.maths.unsw.edu.au/~rsw/Sphere/Points/SF/SF29-Nov-2012/"
    fileurl = url + filename

    http = urllib3.PoolManager(
        cert_reqs=False)
    http_data = http.urlopen('GET', fileurl)

    if http_data.status == 200:
        file_data = http_data.data.decode()
    elif http_data.status == 404:
        raise FileNotFoundError("File was not found. Check if the design you \
                are trying to calculate is a valid t-design.")
    else:
        raise ConnectionError("Connection error. Please check your internet \
                connection.")

    points = np.fromstring(
        file_data,
        dtype=float,
        sep=' ').reshape((n_points, 3)).T

    return SamplingSphere.from_array(points)


def dodecahedron():
    """Generate a sampling based on the center points of the twelve
    dodecahedron faces.

    Returns
    -------
    rad : ndarray
        Radius of the sampling points
    theta : ndarray
        Elevation angle in the range [0, pi]
    phi : ndarray
        Azimuth angle in the range [0, 2 pi]
    Returns
    -------
    sampling : SamplingSphere
        SamplingSphere object containing all sampling points
    """

    dihedral = 2*np.arcsin(np.cos(np.pi/3)/np.sin(np.pi/5))
    R = np.tan(np.pi/3)*np.tan(dihedral/2)
    rho = np.cos(np.pi/5)/np.sin(np.pi/10)

    theta1 = np.arccos((np.cos(np.pi/5)/np.sin(np.pi/5))/np.tan(np.pi/3))

    a2 = 2*np.arccos(rho/R)

    theta2 = theta1+a2
    theta3 = np.pi - theta2
    theta4 = np.pi - theta1

    phi1 = 0
    phi2 = 2*np.pi/3
    phi3 = 4*np.pi/3

    theta = np.concatenate((
        np.tile(theta1, 3), np.tile(theta2, 3),
        np.tile(theta3, 3), np.tile(theta4, 3)))
    phi = np.tile(np.array([
        phi1, phi2, phi3,
        phi1 + np.pi/3, phi2 + np.pi/3, phi3 + np.pi/3]), 2)

    rad = np.ones(np.size(theta))

    return SamplingSphere.from_spherical(rad, theta, phi)


def icosahedron():
    """Generate a sampling based on the center points of the twenty \
            icosahedron faces.

    Returns
    -------
    sampling : SamplingSphere
        SamplingSphere object containing all sampling points
    """
    gamma_R_r = np.arccos(np.cos(np.pi/3) / np.sin(np.pi/5))
    gamma_R_rho = np.arccos(1/(np.tan(np.pi/5) * np.tan(np.pi/3)))

    theta = np.tile(np.array([np.pi - gamma_R_rho,
                              np.pi - gamma_R_rho - 2*gamma_R_r,
                              2*gamma_R_r + gamma_R_rho,
                              gamma_R_rho]), 5)
    theta = np.sort(theta)
    phi = np.arange(0, 2*np.pi, 2*np.pi/5)
    phi = np.concatenate((np.tile(phi, 2), np.tile(phi + np.pi/5, 2)))
    rad = np.ones(20)

    return SamplingSphere.from_spherical(rad, theta, phi)


def equiangular(n_max):
    """Generate an equiangular sampling of the sphere.

    Paramters
    ---------
    n_max : integer
        Spherical harmonic order of the sampling

    Returns
    -------
    sampling : SamplingSphere
        SamplingSphere object containing all sampling points

    """
    n_theta = np.round((n_max+1)*2)
    n_phi = n_theta
    theta_angles = np.arange(np.pi/(n_theta*2),
                             np.pi,
                             np.pi/n_theta)
    phi_angles = np.arange(0,
                           2*np.pi,
                           2*np.pi/n_phi)
    theta, phi = np.meshgrid(theta_angles, phi_angles)
    rad = np.ones(theta.size)

    # calculate weights
    L = 2*np.arange(0, n_max + 1) + 1
    factor_phi = 2*np.pi/n_phi
    factor_theta = 2/n_theta
    factor_sin = 4/np.pi * np.sin(theta_angles) * \
        (1/L @ L[np.newaxis].T @ theta_angles[np.newaxis])
    weights = np.tile(factor_phi * factor_theta * np.pi/2 * factor_sin, n_phi)

    sampling = SamplingSphere.from_spherical(
        rad, theta.reshape(-1), phi.reshape(-1))
    sampling.weights = weights

    return sampling


def gaussian(n_max):
    """Generate sampling of the sphere based on the Gaussian quadrature.

    Paramters
    ---------
    n_max : integer
        Spherical harmonic order of the sampling

    Returns
    -------
    sampling : SamplingSphere
        SamplingSphere object containing all sampling points

    """
    legendre, weights = np.polynomial.legendre.leggauss(n_max+1)
    theta_angles = np.arccos(legendre)
    n_phi = np.round((n_max+1)*2)
    phi_angles = np.arange(0,
                           2*np.pi,
                           2*np.pi/n_phi)
    theta, phi = np.meshgrid(theta_angles, phi_angles)
    rad = np.ones(theta.size)
    weights = np.tile(weights*np.pi/(n_max+1), 2*(n_max+1))

    sampling = SamplingSphere.from_spherical(
        rad, theta.reshape(-1), phi.reshape(-1))
    sampling.weights = weights
    return sampling


def eigenmike_em32():
    """Microphone positions of the Eigenmike em32 by mhacoustics according to
    the Eigenstudio user manual on the homepage [6]_.

    References
    ----------
    .. [6]  Eigenstudio User Manual, https://mhacoustics.com/download


    Returns
    -------
    sampling : SamplingSphere
        SamplingSphere object containing all sampling points

    """
    rad = np.ones(32)
    theta = np.array([69.0, 90.0, 111.0, 90.0, 32.0, 55.0,
                      90.0, 125.0, 148.0, 125.0, 90.0, 55.0,
                      21.0, 58.0, 121.0, 159.0, 69.0, 90.0,
                      111.0, 90.0, 32.0, 55.0, 90.0, 125.0,
                      148.0, 125.0, 90.0, 55.0, 21.0, 58.0,
                      122.0, 159.0]) * np.pi / 180
    phi = np.array([0.0, 32.0, 0.0, 328.0, 0.0, 45.0,
                    69.0, 45.0, 0, 315.0, 291.0, 315.0,
                    91.0, 90.0, 90.0, 89.0, 180.0, 212.0,
                    180.0, 148.0, 180.0, 225.0, 249.0, 225.0,
                    180.0, 135.0, 111.0, 135.0, 269.0, 270.0,
                    270.0, 271.0]) * np.pi / 180

    return SamplingSphere.from_spherical(rad, theta, phi)


def icosahedron_ke4():
    """Microphone positions of the KE4 spherical microphone array.
    The microphone marked as "1" defines the positive x-axis.

    Returns
    -------
    sampling : SamplingSphere
        SamplingSphere object containing all sampling points

    """

    theta = np.array([1.565269801525254, 2.294997457752220, 1.226592351686568,
                      1.226592351686568, 1.696605844149543, 2.409088979442091,
                      2.409088979442090, 1.696605844149543, 0.506378251408914,
                      0.506378251408914, 2.635214402180879, 1.444986809440250,
                      0.732503674147703, 0.732503674147703, 1.444986809440250,
                      2.635214402180879, 1.915000301903226, 0.846595195837573,
                      1.915000301903225, 1.576322852064539])

    phi = np.array([0.000000000000000, 6.283185307179586, 0.660263824203969,
                    5.622921482975618, 5.055791576545843, 5.241315660913181,
                    1.041869646266405, 1.227393730633743, 0.826693168093822,
                    5.456492139085764, 3.968285821683615, 4.368986384223536,
                    4.183462299856198, 2.099723007323388, 1.914198922956050,
                    2.314899485495971, 2.481328829385824, 3.141592653589793,
                    3.801856477793762, 3.141592653589793])

    rad = np.ones(20) * 0.065

    return SamplingSphere.from_spherical(rad, theta, phi)


def equalarea(n_max, condition_num=2.5, n_points=None):
    """Sampling based on partitioning into faces with equal area [9]_.

    Parameters
    ----------
    n_max : int
        Spherical harmonic order
    condition_num : double
        Desired maximum condition number of the spherical harmonic basis matrix
    n_points : int, optional
        Number of points to start the condition number optimization. If set to
        None n_points will be (n_max+1)**2

    Returns
    -------
    sampling : SamplingSphere
        SamplingSphere object containing all sampling points

    References
    ----------
    .. [9]  P. Leopardi, “A partition of the unit sphere into regions of equal
            area and small diameter,” Electronic Transactions on Numerical
            Analysis, vol. 25, no. 12, pp. 309–327, 2006.

    """
    if not n_points:
        n_points = (n_max+1)**2

    while True:
        point_set = eq_point_set(2, n_points)
        sampling = SamplingSphere(point_set[0], point_set[1], point_set[2])

        if condition_num == np.inf:
            break
        Y = spharpy.spherical.spherical_harmonic_basis(n_max, sampling)
        cond = np.linalg.cond(Y)
        if cond < condition_num:
            break
        n_points += 1

    sampling.n_max = n_max
    return sampling


def spiral_points(n_max, condition_num=2.5, n_points=None):
    """Sampling based on a spiral distribution of points on a sphere [10]_.

    Parameters
    ----------
    n_max : int
        Spherical harmonic order
    condition_num : double
        Desired maximum condition number of the spherical harmonic basis matrix
    n_points : int, optional
        Number of points to start the condition number optimization. If set to
        None n_points will be (n_max+1)**2

    Returns
    -------
    sampling : SamplingSphere
        SamplingSphere object containing all sampling points

    References
    ----------

    .. [10]  E. a. Rakhmanov, E. B. Saff, and Y. M. Zhou, “Minimal Discrete
            Energy on the Sphere,” Mathematical Research Letters, vol. 1,
            no. 6, pp. 647–662, 1994.

    """
    if n_points is None:
        n_points = (n_max+1)**2

    def _spiral_points(n_points):
        """Helper function doing the actual calculation of the points"""
        r = np.zeros(n_points)
        h = np.zeros(n_points)
        theta = np.zeros(n_points)
        phi = np.zeros(n_points)

        p = 1/2
        a = 1 - 2*p/(n_points-3)
        b = p*(n_points+1)/(n_points-3)
        r[0] = 0
        theta[0] = np.pi
        phi[0] = 0
        # Then for k stepping by 1 from 2 to n-1:
        for k in range(1, n_points-1):
            kStrich = a*k + b
            h[k] = -1 + 2*(kStrich-1)/(n_points-1)
            r[k] = np.sqrt(1-h[k]**2)
            theta[k] = np.arccos(h[k])
            phi[k] = np.mod(
                (phi[k-1]) + 3.6/np.sqrt(n_points)*2/(r[k-1]+r[k]), 2*np.pi)
        # Finally:
        theta[n_points-1] = 0
        phi[n_points-1] = 0

        return theta, phi

    while True:
        theta, phi = _spiral_points(n_points)
        sampling = SamplingSphere.from_spherical(np.ones(n_points), theta, phi)
        if condition_num == np.inf:
            break
        Y = spharpy.spherical.spherical_harmonic_basis(n_max, sampling)
        cond = np.linalg.cond(Y)
        if cond < condition_num:
            break
        n_points += 1

    sampling.n_max = n_max

    return sampling



"""
Tests for special functions
"""

import pytest
from spharpy import special
from spharpy import samplings

import numpy as np
import numpy.testing as npt


def genfromtxt_complex(filename, delimiter=','):
    """generate complex numpy array from csv file."""
    data_str = np.genfromtxt(filename, delimiter=delimiter, dtype=str)
    mapping = np.vectorize(lambda t: complex(t.replace('i', 'j')))
    return mapping(data_str)


class TestBessel(object):
    def test_shape(self):
        n = np.array([0, 1])
        z = np.linspace(0.1, 5, 10)
        res = special.spherical_bessel(n, z)

        shape = (2, 10)
        assert shape == res.shape

    def test_val(self):
        z = np.linspace(0, 10, 25)
        n = [0, 1, 2]
        res = special.spherical_bessel(n, z)
        truth = np.genfromtxt('./tests/data/bessel.csv', delimiter=',')
        npt.assert_allclose(res, truth)


class TestBesselPrime(object):
    def test_shape(self):
        n = np.array([0, 1])
        z = np.linspace(0.1, 5, 10)
        res = special.spherical_bessel(n, z, derivative=True)

        shape = (2, 10)
        assert shape == res.shape

    def test_val(self):
        z = np.linspace(0.1, 10, 25)
        n = [0, 1, 2]
        res = special.spherical_bessel(n, z, derivative=True)
        truth = np.genfromtxt('./tests/data/bessel_diff.csv', delimiter=',')
        npt.assert_allclose(res, truth)


class TestHankel(object):
    def test_shape(self):
        n = np.array([0, 1])
        z = np.linspace(0.1, 5, 10)
        res = special.spherical_hankel(n, z, kind=1)

        shape = (2, 10)
        assert shape == res.shape

    def test_kind_exception(self):
        with pytest.raises(ValueError):
            special.spherical_hankel([0], [1], kind=3)

    def test_val_second_kind(self):
        z = np.linspace(0.1, 5, 25)
        n = np.array([0, 1, 2])
        res = special.spherical_hankel(n, z, kind=2)
        truth = genfromtxt_complex('./tests/data/hankel_2.csv', delimiter=',')
        npt.assert_allclose(res, truth)

    def test_val_first_kind(self):
        z = np.linspace(0.1, 5, 25)
        n = np.array([0, 1, 2])
        res = special.spherical_hankel(n, z, kind=1)
        truth = genfromtxt_complex('./tests/data/hankel_1.csv', delimiter=',')
        npt.assert_allclose(res, truth)


class TestHankelPrime(object):
    def test_shape(self):
        n = np.array([0, 1])
        z = np.linspace(0.1, 5, 10)
        res = special.spherical_hankel(n, z, kind=1, derivative=True)

        shape = (2, 10)
        assert shape == res.shape

    def test_kind_exception(self):
        with pytest.raises(ValueError):
            special.spherical_hankel([0], [1], kind=3, derivative=True)

    def test_val_second_kind(self):
        z = np.linspace(0.1, 5, 25)
        n = [0, 1, 2]
        res = special.spherical_hankel(n, z, kind=2, derivative=True)
        truth = genfromtxt_complex('./tests/data/hankel_2_diff.csv', delimiter=',')
        npt.assert_allclose(res, truth)

    def test_val_first_kind(self):
        z = np.linspace(0.1, 5, 25)
        n = [0, 1, 2]
        res = special.spherical_hankel(n, z, kind=1, derivative=True)
        truth = genfromtxt_complex('./tests/data/hankel_1_diff.csv', delimiter=',')
        npt.assert_allclose(res, truth)


def test_spherical_harmonic_derivative_theta():
    n_max = 5
    theta = np.array([np.pi/2, np.pi/2, 0, np.pi/2, np.pi/4])
    phi = np.array([0, np.pi/2, 0, np.pi/4, np.pi/4])

    desired_all = np.genfromtxt(
        './tests/data/Y_grad_ele.csv',
        dtype=complex,
        delimiter=',')

    for acn in range((n_max+1)**2):
        n = int((np.ceil(np.sqrt(acn + 1)) - 1))
        m = int(acn - n**2 - n)

        actual = special.spherical_harmonic_derivative_theta(n, m, theta, phi)
        desired = desired_all[:, acn]

        npt.assert_allclose(actual, desired, rtol=1e-10, atol=1e-10)


def test_spherical_harmonic_derivative_phi():
    n_max = 5
    theta = np.array([np.pi/2, np.pi/2, 0, np.pi/2, np.pi/4])
    phi = np.array([0, np.pi/2, 0, np.pi/4, np.pi/4])

    desired_all = np.genfromtxt(
        './tests/data/Y_diff_azi.csv',
        dtype=complex,
        delimiter=',')

    for acn in range((n_max+1)**2):
        n = int((np.ceil(np.sqrt(acn + 1)) - 1))
        m = int(acn - n**2 - n)

        actual = special.spherical_harmonic_derivative_phi(n, m, theta, phi)
        desired = desired_all[:, acn]

        npt.assert_allclose(actual, desired, rtol=1e-10, atol=1e-10)


def test_spherical_harmonic_gradient_phi():
    n_max = 5
    theta = np.array([np.pi/2, np.pi/2, 0, np.pi/2, np.pi/4])
    phi = np.array([0, np.pi/2, 0, np.pi/4, np.pi/4])

    desired_all = np.genfromtxt(
        './tests/data/Y_grad_azi.csv',
        dtype=complex,
        delimiter=',')

    for acn in range((n_max+1)**2):
        n = int((np.ceil(np.sqrt(acn + 1)) - 1))
        m = int(acn - n**2 - n)

        actual = special.spherical_harmonic_gradient_phi(n, m, theta, phi)
        desired = desired_all[:, acn]

        npt.assert_allclose(actual, desired, rtol=1e-10, atol=1e-10)


def test_spherical_harmonic_derivative_theta_real():
    n_max = 5
    theta = np.array([np.pi/2, np.pi/2, 0, np.pi/2, np.pi/4])
    phi = np.array([0, np.pi/2, 0, np.pi/4, np.pi/4])

    desired_all = np.genfromtxt(
        './tests/data/Y_grad_real_ele.csv',
        dtype=float,
        delimiter=',')

    for acn in range((n_max+1)**2):
        n = int((np.ceil(np.sqrt(acn + 1)) - 1))
        m = int(acn - n**2 - n)

        actual = special.spherical_harmonic_derivative_theta_real(
            n, m, theta, phi)
        desired = desired_all[:, acn]

        npt.assert_allclose(actual, desired, rtol=1e-10, atol=1e-10)


def test_spherical_harmonic_derivative_phi_real():
    n_max = 5
    theta = np.array([np.pi/2, np.pi/2, 0, np.pi/2, np.pi/4])
    phi = np.array([0, np.pi/2, 0, np.pi/4, np.pi/4])

    desired_all = np.genfromtxt(
        './tests/data/Y_diff_real_azi.csv',
        dtype=float,
        delimiter=',')

    for acn in range((n_max+1)**2):
        n = int((np.ceil(np.sqrt(acn + 1)) - 1))
        m = int(acn - n**2 - n)

        actual = special.spherical_harmonic_derivative_phi_real(
            n, m, theta, phi)
        desired = desired_all[:, acn]

        npt.assert_allclose(actual, desired, rtol=1e-10, atol=1e-10)


def test_spherical_harmonic_gradient_phi_real():
    n_max = 5
    theta = np.array([np.pi/2, np.pi/2, 0, np.pi/2, np.pi/4])
    phi = np.array([0, np.pi/2, 0, np.pi/4, np.pi/4])

    desired_all = np.genfromtxt(
        './tests/data/Y_grad_real_azi.csv',
        dtype=float,
        delimiter=',')

    for acn in range((n_max+1)**2):
        n = int((np.ceil(np.sqrt(acn + 1)) - 1))
        m = int(acn - n**2 - n)

        actual = special.spherical_harmonic_gradient_phi_real(n, m, theta, phi)
        desired = desired_all[:, acn]

        npt.assert_allclose(actual, desired, rtol=1e-10, atol=1e-10)



"""
Plot functions for spatial data
"""

import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.tri as mtri
import numpy as np
import scipy.spatial as sspat
from matplotlib import cm, colors
from mpl_toolkits.mplot3d import Axes3D
__all__ = [Axes3D]
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from packaging import version
from scipy.stats import circmean

from .cmap import phase_twilight

from spharpy.samplings import sph2cart, spherical_voronoi


def set_aspect_equal_3d(ax):
    """Fix equal aspect bug for 3D plots."""

    xlim = ax.get_xlim3d()
    ylim = ax.get_ylim3d()
    zlim = ax.get_zlim3d()

    from numpy import mean
    xmean = mean(xlim)
    ymean = mean(ylim)
    zmean = mean(zlim)

    plot_radius = max([abs(lim - mean_)
                       for lims, mean_ in ((xlim, xmean),
                                           (ylim, ymean),
                                           (zlim, zmean))
                       for lim in lims])

    ax.set_xlim3d([xmean - plot_radius, xmean + plot_radius])
    ax.set_ylim3d([ymean - plot_radius, ymean + plot_radius])
    ax.set_zlim3d([zmean - plot_radius, zmean + plot_radius])


def scatter(coordinates, ax=None):
    """Plot the x, y, and z coordinates of the sampling grid in the 3d space.

    Parameters
    ----------
    coordinates : Coordinates

    """
    fig = plt.gcf()
    if ax is None:
        ax = plt.gca() if fig.axes else plt.axes(projection='3d')

    if '3d' not in ax.name:
        raise ValueError("The projection of the axis needs to be '3d'")

    ax.scatter(coordinates.x, coordinates.y, coordinates.z)
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')

    ax.set_box_aspect([
        np.ptp(coordinates.x),
        np.ptp(coordinates.y),
        np.ptp(coordinates.z)])

    plt.show()


def _triangulation_sphere(sampling, data):
    """Triangulation for data points sampled on a spherical surface.

    Parameters
    ----------
    sampling : Coordinates
        Coordinate object for which the triangulation is calculated
    xyz : list of arrays
        x, y, and z values of the data points in the triangulation

    Returns
    -------
    triangulation : matplotlib Triangulation

    """
    x, y, z = sph2cart(
        np.abs(data),
        sampling.elevation,
        sampling.azimuth)
    hull = sspat.ConvexHull(
        np.asarray(sph2cart(
            np.ones(sampling.n_points),
            sampling.elevation,
            sampling.azimuth)).T)
    tri = mtri.Triangulation(x, y, triangles=hull.simplices)

    return tri, [x, y, z]


def interpolate_data_on_sphere(
        sampling,
        data,
        overlap=np.pi*0.25,
        refine=False,
        interpolator='linear'):
    """Linear interpolator for data on a spherical surface. The interpolator
    exploits that the data on the sphere is periodic with regard to the
    elevation and azimuth angle. The data is periodically extended to a
    specified overlap angle before interpolation.

    Parameters
    ----------
    sampling : Coordinates
        The coordinates at which the data is sampled.
    data : ndarray, double
        The sampled data points.
    overlap : float, (pi/4)
        The overlap for the periodic extension in azimuth angle, given in
        radians
    refine : bool (False)
        Refine the mesh before interpolating
    interpolator : linear, cubic
        The interpolation method to be used

    Returns
    -------
    interp : LinearTriInterpolator, CubicTriInterpolator
        The interpolator object.

    Note
    ----
    Internally, matplotlibs LinearTriInterpolator or CubicTriInterpolator
    are used.

    """
    lats = sampling.latitude
    lons = sampling.longitude

    mask = lons > np.pi - overlap
    lons = np.concatenate((lons, lons[mask] - np.pi*2))
    lats = np.concatenate((lats, lats[mask]))
    data = np.concatenate((data, data[mask]))

    mask = lons < -np.pi + overlap
    lons = np.concatenate((lons, lons[mask] + np.pi*2))
    lats = np.concatenate((lats, lats[mask]))
    data = np.concatenate((data, data[mask]))

    tri = mtri.Triangulation(lons, lats)

    if refine:
        refiner = mtri.UniformTriRefiner(tri)
        tri, data = refiner.refine_field(
            data,
            triinterpolator=mtri.LinearTriInterpolator(tri, data),
            subdiv=3)

    if interpolator == 'linear':
        interpolator = mtri.LinearTriInterpolator(tri, data)
    elif interpolator == 'cubic':
        interpolator = mtri.CubicTriInterpolator(tri, data, kind='mind_E')
    else:
        raise ValueError("Please give a valid interpolation method.")

    return interpolator


def _balloon_color_data(tri, data, itype):
    """Return the data array that is to be mapped to the colormap of the
    balloon.

    Parameters
    ----------
    tri : Triangulation
        The matplotlib triangulation for the sphere
    data : ndarray, double, complex double
        The data array
    itype : 'magnitude', 'phase', 'amplitude'
        Whether to plot magnitude levels or the phase.

    Returns
    -------
    color_data : ndarray, double
        The data array for the colormap.
    vmin : double
        The minimum of the color data

    vmax : double
        The maximum of the color data


    """
    if itype == 'phase':
        cdata = np.mod(np.angle(data), 2*np.pi)
        vmin = 0
        vmax = 2*np.pi
        colors = circmean(cdata[tri.triangles], axis=1)
    elif itype == 'magnitude':
        cdata = np.abs(data)
        vmin = np.min(cdata)
        vmax = np.max(cdata)
        colors = np.mean(cdata[tri.triangles], axis=1)
    elif itype == 'amplitude':
        vmin = np.min(data)
        vmax = np.max(data)
        colors = np.mean(data[tri.triangles], axis=1)
    else:
        raise ValueError("Invalid type of data mapping.")

    return colors, vmin, vmax


def pcolor_sphere(
        coordinates,
        data,
        cmap=None,
        colorbar=True,
        show=True,
        phase=False,
        ax=None,
        *args,
        **kwargs):
    """Plot data on the surface of a sphere defined by the coordinate angles
    theta and phi. The data array will be mapped onto the surface of a sphere.

    Note
    ----
    When plotting the phase encoded in the colormap, the function will switch
    to the HSV colormap and ignore the user input for the cmap input variable.

    Parameters
    ----------
    coordinates : Coordinates
        Coordinates defining a sphere
    data : ndarray, double
        Data for each angle, must have size corresponding to the number of
        points given in coordinates.
    cmap : matplotlib colomap, optional
        Colormap for the plot, see matplotlib.cm
    phase : boolean, optional
        Encode the phase of the data in the colormap. This option will be
        activated by default of the data is complex valued.
    ax : matplotlib.axis, None, optional
        The matplotlib axis object used for plotting. By default `None`, which
        will create a new axis object.
    show : boolean, optional
        Whether to show the figure or not

    """
    tri, xyz = _triangulation_sphere(coordinates, np.ones_like(data))
    fig = plt.gcf()

    if ax is None:
        ax = plt.gca() if fig.axes else plt.axes(projection='3d')

    elif '3d' not in ax.name:
        raise ValueError("The projection of the axis needs to be '3d'")

    if np.iscomplex(data).any() or phase:
        itype = 'phase'
        if cmap is None:
            cmap = phase_twilight()
        clabel = 'Phase (rad)'
    else:
        itype = 'amplitude'
        if cmap is None:
            cmap = cm.viridis
        clabel = 'Amplitude'

    cdata, vmin, vmax = _balloon_color_data(tri, data, itype)

    plot = ax.plot_trisurf(tri,
                           xyz[2],
                           cmap=cmap,
                           antialiased=True,
                           vmin=vmin,
                           vmax=vmax)

    plot.set_array(cdata)

    if colorbar:
        fig.colorbar(plot, ax=ax, label=clabel)

    ax.set_xlabel('x[m]')
    ax.set_ylabel('y[m]')
    ax.set_zlabel('z[m]')

    ax.set_box_aspect([
        np.ptp(coordinates.x),
        np.ptp(coordinates.y),
        np.ptp(coordinates.z)])

    if show:
        plt.show()

    return plot


def balloon_wireframe(
        coordinates,
        data,
        cmap=None,
        phase=False,
        show=True,
        colorbar=True,
        ax=None):
    """Plot data on a sphere defined by the coordinate angles
    theta and phi. The magnitude information is mapped onto the radius of the
    sphere. The colormap represents either the phase or the magnitude of the
    data array.

    Note
    ----
    When plotting the phase encoded in the colormap, the function will switch
    to the HSV colormap and ignore the user input for the cmap input variable.

    Parameters
    ----------
    coordinates : Coordinates
        Coordinates defining a sphere
    data : ndarray, double
        Data for each angle, must have size corresponding to the number of
        points given in coordinates.
    cmap : matplotlib colomap, optional
        Colormap for the plot, see matplotlib.cm
    phase : boolean, optional
        Encode the phase of the data in the colormap. This option will be
        activated by default of the data is complex valued.
    show : boolean, optional
        Whether to show the figure or not
    """
    tri, xyz = _triangulation_sphere(coordinates, data)
    fig = plt.gcf()

    if ax is None:
        ax = plt.gca() if fig.axes else plt.axes(projection='3d')

    elif '3d' not in ax.name:
        raise ValueError("The projection of the axis needs to be '3d'")

    if np.iscomplex(data).any() or phase:
        itype = 'phase'
        if cmap is None:
            cmap = phase_twilight()
        clabel = 'Phase (rad)'
    else:
        itype = 'amplitude'
        if cmap is None:
            cmap = cm.viridis
        clabel = 'Amplitude'

    cdata, vmin, vmax = _balloon_color_data(tri, data, itype)

    plot = ax.plot_trisurf(tri,
                           xyz[2],
                           antialiased=True,
                           vmin=vmin,
                           vmax=vmax)

    cnorm = plt.Normalize(vmin, vmax)
    cmap_colors = cmap(cnorm(cdata))

    cmappable = mpl.cm.ScalarMappable(cnorm, cmap)
    cmappable.set_array(np.linspace(vmin, vmax, cdata.size))

    plot.set_edgecolors(cmap_colors)
    plot.set_facecolors(np.ones(cmap_colors.shape)*0.9)

    if colorbar:
        fig.colorbar(cmappable, ax=ax, label=clabel)

    ax.set_xlabel('x[m]')
    ax.set_ylabel('y[m]')
    ax.set_zlabel('z[m]')

    plot.set_facecolors(np.ones(cmap_colors.shape)*0.9)

    ax.set_box_aspect([
        np.ptp(xyz[0]),
        np.ptp(xyz[1]),
        np.ptp(xyz[2])])

    if show:
        plt.show()

    plot.set_facecolor([0.9, 0.9, 0.9, 0.9])

    return plot


def balloon(
        coordinates,
        data,
        cmap=None,
        phase=False,
        show=True,
        colorbar=True,
        ax=None,
        *args,
        **kwargs):
    """Plot data on a sphere defined by the coordinate angles theta and phi.
    The magnitude information is mapped onto the radius of the sphere.
    The colormap represents either the phase or the magnitude of the
    data array.


    Note
    ----
    When plotting the phase encoded in the colormap, the function will switch
    to the HSV colormap and ignore the user input for the cmap input variable.

    Parameters
    ----------
    coordinates : Coordinates
        Coordinates defining a sphere
    data : ndarray, double
        Data for each angle, must have size corresponding to the number of
        points given in coordinates.
    cmap : matplotlib colomap, optional
        Colormap for the plot, see matplotlib.cm
    phase : boolean, optional
        Encode the phase of the data in the colormap. This option will be
        activated by default of the data is complex valued.
    show : boolean, optional
        Wheter to show the figure or not
    """
    tri, xyz = _triangulation_sphere(coordinates, data)
    fig = plt.gcf()

    if ax is None:
        ax = plt.gca() if fig.axes else plt.axes(projection='3d')

    elif '3d' not in ax.name:
        raise ValueError("The projection of the axis needs to be '3d'")

    if np.iscomplex(data).any() or phase:
        itype = 'phase'
        if cmap is None:
            cmap = phase_twilight()
        clabel = 'Phase (rad)'
    else:
        itype = 'amplitude'
        if cmap is None:
            cmap = cm.viridis
        clabel = 'Amplitude'

    cdata, vmin, vmax = _balloon_color_data(tri, data, itype)

    plot = ax.plot_trisurf(tri,
                           xyz[2],
                           cmap=cmap,
                           antialiased=True,
                           vmin=vmin,
                           vmax=vmax,
                           *args,
                           **kwargs)

    plot.set_array(cdata)

    ax.set_box_aspect([
        np.ptp(xyz[0]),
        np.ptp(xyz[1]),
        np.ptp(xyz[2])])

    if colorbar:
        fig.colorbar(plot, ax=ax, label=clabel)

    ax.set_xlabel('x[m]')
    ax.set_ylabel('y[m]')
    ax.set_zlabel('z[m]')

    if show:
        plt.show()

    return plot


def voronoi_cells_sphere(sampling, round_decimals=13, ax=None):
    """Plot the Voronoi cells of a Voronoi tesselation on a sphere.

    Parameters
    ----------
    sampling : SamplingSphere
        Sampling as SamplingSphere object
    round_decimals : int
        Decimals to be rounded to for eliminating duplicate points in
        the voronoi diagram
    ax : AxesSubplot, None, optional
        The subplot axes to use for plotting. The used projection needs to be
        '3d'.

    """
    sv = spherical_voronoi(sampling, round_decimals=round_decimals)
    sv.sort_vertices_of_regions()
    points = sampling.cartesian.T

    fig = plt.gcf()
    if ax is None:
        ax = plt.gca() if fig.axes else plt.axes(projection='3d')

    if '3d' not in ax.name:
        raise ValueError("The projection of the axis needs to be '3d'")

    if version.parse(mpl.__version__) < version.parse('3.1.0'):
        ax.set_aspect('equal')

    # plot the unit sphere for reference (optional)
    u = np.linspace(0, 2 * np.pi, 100)
    v = np.linspace(0, np.pi, 100)
    x = np.outer(np.cos(u), np.sin(v))
    y = np.outer(np.sin(u), np.sin(v))
    z = np.outer(np.ones(np.size(u)), np.cos(v))
    ax.plot_surface(x, y, z, color='y', alpha=0.1)

    ax.scatter(points[:, 0], points[:, 1], points[:, 2], c='r')

    for region in sv.regions:
        polygon = Poly3DCollection(
            [sv.vertices[region]], alpha=0.5, facecolor=None)
        polygon.set_edgecolor((0, 0, 0, 1))
        polygon.set_facecolor((1, 1, 1, 0.))

        ax.add_collection3d(polygon)

    ax.set_box_aspect([
        np.ptp(sampling.x),
        np.ptp(sampling.y),
        np.ptp(sampling.z)])

    ax.set_xlabel('x[m]')
    ax.set_ylabel('y[m]')
    ax.set_zlabel('z[m]')


def _combined_contour(x, y, data, limits, cmap, ax):
    """Combine a filled contour plot with a black line contour plot for
    better highlighting.

    Parameters
    ----------
    x : ndarray, double
        The x coordinates.
    y : ndarray, double
        The y coordinates.
    data : ndarray, double
        The data array.
    limits : tuple, list
        Tuple or list containing the maximum and minimum to which the colormap
        needs to be clipped.
    cmap : matplotlib.colormap
        The colormap
    ax : matplotlib.axes
        The axes object into which the contour is plotted

    Returns
    -------
    cf : matplotlib.ScalarMappable
        The scalar mappable for the contour plot

    """
    extend = 'neither'
    if limits is None:
        limits = (data.min(), data.max())
    else:
        mask_min = data < limits[0]
        data[mask_min] = limits[0]
        mask_max = data > limits[1]
        data[mask_max] = limits[1]
        if np.any(mask_max) & np.any(mask_min):
            extend = 'both'
        elif np.any(mask_max) & ~np.any(mask_min):
            extend = 'max'
        elif ~np.any(mask_max) & np.any(mask_min):
            extend = 'min'

    ax.tricontour(x, y, data, linewidths=0.5, colors='k',
                  vmin=limits[0], vmax=limits[1], extend=extend)
    return ax.tricontourf(
        x, y, data, cmap=cmap, vmin=limits[0], vmax=limits[1], extend=extend)


def pcolor_map(
        coordinates,
        data,
        projection='mollweide',
        limits=None,
        cmap=plt.get_cmap('viridis'),
        show=True,
        refine=False,
        ax=None,
        **kwargs):
    """
    Plot the map projection of data points sampled on a spherical surface.
    The data has to be real.

    Notes
    -----
    In case limits are given, all out of bounds data will be clipped to the
    respective limit.

    Parameters
    ----------
    latitude: ndarray, double
        Geodetic latitude angle of the map, must be in [-pi/2, pi/2]
    longitude: ndarray, double
        Geodetic longitude angle of the map, must be in [-pi, pi]
    data: ndarray, double
        Data for each angle, must have size corresponding to the number of
        points given in coordinates.
    show : boolean, optional
        Wheter to show the figure or not

    """
    tri = mtri.Triangulation(coordinates.longitude, coordinates.latitude)
    if refine is not None:
        subdiv = refine if isinstance(refine, int) else 2
        refiner = mtri.UniformTriRefiner(tri)
        tri, data = refiner.refine_field(
            data,
            triinterpolator=mtri.LinearTriInterpolator(tri, data),
            subdiv=subdiv)

    fig = plt.gcf()

    if ax is None:
        ax = plt.gca() if fig.axes else plt.axes(projection=projection)

    if ax.name != projection:
        raise ValueError(
            "Projection does not match the projection of the axis",
            f"Needs to be '{projection}', but is '{ax.name}'")

    ax.set_xlabel('Longitude [$^\\circ$]')
    ax.set_ylabel('Latitude [$^\\circ$]')

    extend = 'neither'
    if limits is None:
        limits = (data.min(), data.max())
    else:
        mask_min = data < limits[0]
        data[mask_min] = limits[0]
        mask_max = data > limits[1]
        data[mask_max] = limits[1]
        if np.any(mask_max) & np.any(mask_min):
            extend = 'both'
        elif np.any(mask_max) & ~np.any(mask_min):
            extend = 'max'
        elif ~np.any(mask_max) & np.any(mask_min):
            extend = 'min'

    cf = ax.tripcolor(
        tri, data, cmap=cmap, vmin=limits[0], vmax=limits[1], **kwargs)

    plt.grid(True)
    cb = fig.colorbar(cf, ax=ax, extend=extend)
    cb.set_label('Amplitude')
    if show:
        plt.show()

    return cf


def contour_map(
        coordinates,
        data,
        projection='mollweide',
        limits=None,
        cmap=cm.viridis,
        colorbar=True,
        show=True,
        levels=None,
        ax=None):
    """
    Plot the map projection of data points sampled on a spherical surface.
    The data has to be real.

    Notes
    -----
    In case limits are given, all out of bounds data will be clipped to the
    respective limit.

    Parameters
    ----------
    latitude: ndarray, double
        Geodetic latitude angle of the map, must be in [-pi/2, pi/2]
    longitude: ndarray, double
        Geodetic longitude angle of the map, must be in [-pi, pi]
    data: ndarray, double
        Data for each angle, must have size corresponding to the number of
        points given in coordinates.
    show : boolean, optional
        Wheter to show the figure or not

    """
    fig = plt.gcf()

    res = int(np.ceil(np.sqrt(coordinates.n_points)))

    xi, yi = np.meshgrid(
        np.linspace(-np.pi, np.pi, res*2),
        np.linspace(-np.pi/2, np.pi/2, res))

    interp = interpolate_data_on_sphere(coordinates, data)
    zi = interp(xi, yi)

    if ax is None:
        ax = plt.gca() if fig.axes else plt.axes(projection=projection)

    if ax.name != projection:
        raise ValueError(
            "Projection does not match the projection of the axis",
            f"Needs to be '{projection}', but is '{ax.name}'")

    ax.set_xlabel('Longitude [$^\\circ$]')
    ax.set_ylabel('Latitude [$^\\circ$]')

    extend = 'neither'
    if limits is None:
        limits = (zi.min(), zi.max())
    else:
        mask_min = zi < limits[0]
        zi[mask_min] = limits[0]
        mask_max = zi > limits[1]
        zi[mask_max] = limits[1]
        if np.any(mask_max) and np.any(mask_min):
            extend = 'both'
        elif np.any(mask_max) and not np.any(mask_min):
            extend = 'max'
        elif not np.any(mask_max) and np.any(mask_min):
            extend = 'min'

    ax.contour(xi, yi, zi, levels=levels, linewidths=0.5, colors='k',
               vmin=limits[0], vmax=limits[1], extend=extend)
    cf = ax.pcolormesh(xi, yi, zi, cmap=cmap, shading='gouraud',
                       vmin=limits[0], vmax=limits[1])

    plt.grid(True)
    if colorbar:
        cb = fig.colorbar(cf, ax=ax, ticks=levels)
        cb.set_label('Amplitude')
    if show:
        plt.show()

    return cf


def contour(
        coordinates, data, limits=None, cmap=cm.viridis, show=True, ax=None):
    """
    Plot the map projection of data points sampled on a spherical surface.
    The data has to be real.

    Notes
    -----
    In case limits are given, all out of bounds data will be clipped to the
    respective limit.

    Parameters
    ----------
    latitude: ndarray, double
        Geodetic latitude angle of the map, must be in [-pi/2, pi/2]
    longitude: ndarray, double
        Geodetic longitude angle of the map, must be in [-pi, pi]
    data: ndarray, double
        Data for each angle, must have size corresponding to the number of
        points given in coordinates.
    show : boolean, optional
        Wheter to show the figure or not

    """
    lat_deg = coordinates.latitude * 180/np.pi
    lon_deg = coordinates.longitude * 180/np.pi
    fig = plt.gcf()
    if ax is None:
        ax = plt.gca()
    ax.set_xlabel('Longitude [$^\\circ$]')
    ax.set_ylabel('Latitude [$^\\circ$]')

    cf = _combined_contour(lon_deg, lat_deg, data, limits, cmap, ax)

    plt.grid(True)
    cb = fig.colorbar(cf, ax=ax)
    cb.set_label('Amplitude')
    if show:
        plt.show()

    return cf


class MidpointNormalize(colors.Normalize):
    """Colormap norm with a defined midpoint. Useful for normalization of
    colormaps representing deviations from a defined midpoint.
    Taken from the official matplotlib documentation at
    https://matplotlib.org/users/colormapnorms.html
    """
    def __init__(self, vmin=None, vmax=None, midpoint=0., clip=False):
        self.midpoint = midpoint
        colors.Normalize.__init__(self, vmin, vmax, clip)

    def __call__(self, value, clip=None):
        # I'm ignoring masked values and all kinds of edge cases to make a
        # simple example...
        x, y = [self.vmin, self.midpoint, self.vmax], [0, 0.5, 1]
        return np.ma.masked_array(np.interp(value, x, y))



__author__ = "Marco Berzborn"
__email__ = 'marco.berzborn@akustik.rwth-aachen.de'
__version__ = '0.4.2'

from . import spherical
from . import samplings
from . import plot
from . import indexing
from . import transforms
from . import beamforming
from . import interpolate


__all__ = [
    'spherical',
    'samplings',
    'plot',
    'indexing',
    'transforms',
    'beamforming',
    'interpolate'
]



from scipy import interpolate as spinterpolate
import numpy as np


class SmoothSphereBivariateSpline(spinterpolate.SmoothSphereBivariateSpline):
    """Smooth bivariate spline approximation in spherical coordinates.
    The implementation uses the method proposed by Dierckx [#]_.

    Parameters
    ----------
    sampling : Coordinates
        Coordinates object containing the positions for which the data
        is sampled
    data : array, float
        Array containing the data at the sampling positions. Has to be
        real-valued.
    w : array, float
        Weighting coefficients
    s : float, 1e-4
        Smoothing factor > 0
        Positive smoothing factor defined for estimation condition:
        ``sum((w(i)*(r(i) - s(theta(i), phi(i))))**2, axis=0) <= s``
        Default ``s=len(w)`` which should be a good value if ``1/w[i]`` is
        an estimate of the standard deviation of ``r[i]``. The default
        value is ``1e-4``
    eps : float, 1e-16
        The eps valued to be considered for interpolator estimation.
        Depends on the used data type and numerical precision. The default
        is 1e-16.

    Note
    ----
    This is a wrapper for scipy's SmoothSphereBivariateSpline only using
    Coordinates objects as container for the azimuth and elevation angles.
    For detailed information see scipy's documentation.

    References
    ----------
    .. [#] P. Dierckx, “Algorithms for smoothing data on the sphere with
           tensor product splines,” p. 24, 1984.

    Examples
    --------
    >>> n_max = 10
    >>> sampling = spharpy.samplings.equalarea(
    ...     n_max, n_points=500, condition_num=np.inf)
    >>> Y = spharpy.spherical.spherical_harmonic_basis_real(n_max, sampling)
    >>> y_vec = spharpy.spherical.spherical_harmonic_basis_real(
    ...     n_max, Coordinates(1, 0, 0))
    >>> data = Y @ y_vec.T
    >>> data = np.sin(sampling.elevation)*np.sin(2*sampling.azimuth)
    >>> interp_grid = spharpy.samplings.hyperinterpolation(35)
    >>> data_grid = np.sin(interp_grid.elevation)*np.sin(2*interp_grid.azimuth)
    >>> interpolator = interpolate.SmoothSphereBivariateSpline(
    ...     sampling, data, s=1e-4)
    ...
    >>> interp_data = interpolator(interp_grid)

    """
    def __init__(self, sampling, data, w=None, s=1e-4, eps=1e-16):
        theta = sampling.elevation
        phi = sampling.azimuth
        if np.any(np.iscomplex(data)):
            raise ValueError("Complex data is not supported.")
        super().__init__(theta, phi, data, w, s, eps)

    def __call__(self, interp_grid, dtheta=0, dphi=0):
        """Evaluate the spline on a new sampling grid.

        Parameters
        ----------
        interp_grid : Coordinates
            Coordinates object containing a new set of points for which data
            is to be interpolated.
        dtheta : int, optional
            Order of theta derivative
        dphi : int, optional
            Order of phi derivative

        """
        theta = interp_grid.elevation
        phi = interp_grid.azimuth
        return super().__call__(
            theta, phi, dtheta=dtheta, dphi=dphi, grid=False)

    def get_coeffs(self):
        return super().get_coeffs()

    def get_knots(self):
        return super().get_knots()

    def get_residual(self):
        return super().get_residual()



"""
Subpackage implementing or wrapping special functions required in the
spharpy package.
"""

from itertools import count
import numpy as np
import scipy.special as _spspecial
from scipy.optimize import brentq


def spherical_bessel(n, z, derivative=False):
    r"""
    Spherical bessel function of order n evaluated at z.

    .. math::

        j_n(z) = \sqrt{\frac{\pi}{2z}} J_{n+\frac{1}{2}} (z)

    Parameters
    ----------
    n : int, ndarray
        Order of the spherical bessel function
    z : double, ndarray
        Argument of the spherical bessel function. Has to be real valued.
    derivative : bool
        Return the derivative of the spherical Bessel function


    Returns
    -------
    jn : double, ndarray
        Spherical bessel function. Array with dimensions [N x Z], where N is
        the number of elements in n and Z is the number of elements in z.

    Note
    ----
    This is a wrapper around the Scipy implementation of the spherical Bessel
    function.

    """

    ufunc = _spspecial.spherical_jn
    n = np.asarray(n, dtype=int)
    z = np.asarray(z, dtype=float)

    bessel = np.zeros((n.size, z.size), dtype=complex)

    if n.size > 1:
        for idx, order in zip(count(), n):
            bessel[idx, :] = ufunc(order, z, derivative=derivative)
    else:
        bessel = ufunc(n, z, derivative=derivative)

    if z.ndim <= 1 or n.ndim <= 1:
        bessel = np.squeeze(bessel)

    return bessel


def spherical_bessel_zeros(n_max, n_zeros):
    """Compute the zeros of the spherical Bessel function.
    This function will always start at order zero which is equal
    to sin(x)/x and iteratively compute the roots for higher orders.
    The roots are computed using Brents algorithm from the scipy package.

    Parameters
    ----------
    n_max : int
        The order of the spherical bessel function
    n_zeros : int
        The number of roots to be computed

    Returns
    -------
    roots : ndarray, double
        The roots of the spherical bessel function

    """
    def func(x, n):
        return _spspecial.spherical_jn(n, x)

    zerosj = np.zeros((n_max+1, n_zeros), dtype=float)
    zerosj[0] = np.arange(1, n_zeros+1)*np.pi
    points = np.arange(1, n_zeros+n_max+1)*np.pi

    roots = np.zeros(n_zeros+n_max, dtype=float)
    for i in range(1, n_max+1):
        for j in range(n_zeros+n_max-i):
            roots[j] = brentq(func, points[j], points[j+1], (i,), maxiter=5000)
        points = roots
        zerosj[i, :n_zeros] = roots[:n_zeros]

    return zerosj


def spherical_hankel(n, z, kind=2, derivative=False):
    r"""
    Spherical Hankel function of order n evaluated at z.

    .. math::

        j_n(z) = \sqrt{\frac{\pi}{2z}} J_{n+\frac{1}{2}} (z)

    Parameters
    ----------
    n : int, ndarray
        Order of the spherical bessel function
    z : double, ndarray
        Argument of the spherical bessel function. Has to be real valued.

    Returns
    -------
    hn : double, ndarray
        Spherical bessel function. Array with dimensions [N x Z], where N is
        the number of elements in n and Z is the number of elements in z.

    Note
    ----
    This is based on the Hankel functions implemented in the scipy package.
    """

    if kind not in (1, 2):
        raise ValueError("The spherical hankel function can \
            only be of first or second kind.")

    n = np.asarray(n, dtype=int)
    z = np.asarray(z, dtype=float)

    if derivative:
        ufunc = _spherical_hankel_derivative
    else:
        ufunc = _spherical_hankel

    if n.size > 1:
        hankel = np.zeros((n.size, z.size), dtype=complex)
        for idx, order in zip(count(), n):
            hankel[idx, :] = ufunc(order, z, kind)
    else:
        hankel = ufunc(n, z, kind)

    if z.ndim <= 1 or n.ndim <= 1:
        hankel = np.squeeze(hankel)

    return hankel


def _spherical_hankel(n, z, kind):
    if kind == 1:
        hankel = _spspecial.hankel1(n+0.5, z)
    elif kind == 2:
        hankel = _spspecial.hankel2(n+0.5, z)
    hankel = np.sqrt(np.pi/2/z) * hankel

    return hankel


def _spherical_hankel_derivative(n, z, kind):
    hankel = _spherical_hankel(n-1, z, kind) - \
        (n+1)/z * _spherical_hankel(n, z, kind)

    return hankel


def spherical_harmonic(n, m, theta, phi):
    """The spherical harmonics of order n and degree m.

    n : unsigned int
        The spherical harmonic order
    m : int
        The spherical harmonic degree
    theta : ndarray, double
        The elevation angle
    phi : ndarray, double
        The azimuth angle

    Returns
    -------
    spherical_harmonic : ndarray, double
        The complex valued spherial harmonic of order n and degree m

    Note
    ----
    This function wraps the spherical harmonic implementation from scipy.
    The only difference is that we return zeros instead of nan values
    if $n < |m|$.

    """
    theta = np.asarray(theta, dtype=float)
    phi = np.asarray(phi, dtype=float)

    if n < np.abs(m):
        sph_harm = np.zeros(theta.shape)
    else:
        sph_harm = _spspecial.sph_harm(m, n, phi, theta)
    return sph_harm


def spherical_harmonic_real(n, m, theta, phi):
    r"""Real valued spherical harmonic function of order n and degree m
    evaluated at the angles theta and phi.
    The spherical harmonic functions are fully normalized (N3D) and follow
    the AmbiX phase convention [1]_.

    .. math::

        Y_n^m(\theta, \phi) = \sqrt{\frac{2n+1}{4\pi}
        \frac{(n-|m|)!}{(n+|m|)!}} P_n^{|m|}(\cos \theta)
        \begin{cases}
            \displaystyle \cos(|m|\phi),  & \text{if $m \ge 0$} \newline
            \displaystyle \sin(|m|\phi) ,  & \text{if $m < 0$}
        \end{cases}

    References
    ----------
    .. [1]  C. Nachbar, F. Zotter, E. Deleflie, and A. Sontacchi, “Ambix - A
            Suggested Ambisonics Format (revised by F. Zotter),” International
            Symposium on Ambisonics and Spherical Acoustics,
            vol. 3, pp. 1–11, 2011.



    Parameters
    ----------
    n : unsigned int
        The spherical harmonic order
    m : int
        The spherical harmonic degree
    theta : ndarray, double
        The elevation angle
    phi : ndarray, double
        The azimuth angle

    Returns
    -------
    spherical_harmonic : ndarray, double
        The real valued spherial harmonic of order n and degree m

    """
    # careful here, scipy uses phi as the elevation angle and
    # theta as the azimuth angle
    Y_nm_cplx = _spspecial.sph_harm(m, n, phi, theta)

    if m == 0:
        Y_nm = np.real(Y_nm_cplx)
    elif m > 0:
        Y_nm = np.real(Y_nm_cplx) * np.sqrt(2)
    elif m < 0:
        Y_nm = np.imag(Y_nm_cplx) * np.sqrt(2) * float(-1)**(m+1)

    Y_nm *= float(-1)**(m)

    return Y_nm


def spherical_harmonic_derivative_phi(n, m, theta, phi):
    """Calculate the derivative of the spherical harmonics with respect to
    the azimuth angle phi.

    Parameters
    ----------

    n : int
        Spherical harmonic order
    m : int
        Spherical harmonic degree
    theta : double
        Elevation angle 0 < theta < pi
    phi : double
        Azimuth angle 0 < phi < 2*pi

    Returns
    -------

    sh_diff : complex double
        Spherical harmonic derivative

    """
    if m == 0 or n == 0:
        res = np.zeros(phi.shape, dtype=complex)
    else:
        res = spherical_harmonic(n, m, theta, phi) * 1j * m

    return res


def spherical_harmonic_gradient_phi(n, m, theta, phi):
    """Calculate the derivative of the spherical harmonics with respect to
    the azimuth angle phi divided by sin(theta)

    Parameters
    ----------

    n : int
        Spherical harmonic order
    m : int
        Spherical harmonic degree
    theta : double
        Elevation angle 0 < theta < pi
    phi : double
        Azimuth angle 0 < phi < 2*pi

    Returns
    -------

    sh_diff : complex double
        Spherical harmonic derivative

    """
    if m == 0:
        res = np.zeros(theta.shape, dtype=complex)
    else:
        factor = np.sqrt((2*n+1)/(2*n-1))/2
        exp_phi = np.exp(1j*phi)
        first = np.sqrt((n+m)*(n+m-1)) * exp_phi * \
            spherical_harmonic(n-1, m-1, theta, phi)
        second = np.sqrt((n-m) * (n-m-1)) / exp_phi * \
            spherical_harmonic(n-1, m+1, theta, phi)
        Ynm_sin_theta = (-1) * factor * (first + second)
        res = Ynm_sin_theta * 1j

    return res


def spherical_harmonic_derivative_theta(n, m, theta, phi):
    """Calculate the derivative of the spherical harmonics with respect to
    the elevation angle theta.

    Parameters
    ----------

    n : int
        Spherical harmonic order
    m : int
        Spherical harmonic degree
    theta : double
        Elevation angle 0 < theta < pi
    phi : double
        Azimuth angle 0 < phi < 2*pi

    Returns
    -------

    sh_diff : complex double
        Spherical harmonic derivative

    """
    if n == 0:
        res = np.zeros(theta.shape, dtype=complex)
    else:
        exp_phi = np.exp(1j*phi)
        first = np.sqrt((n-m+1) * (n+m)) * exp_phi * \
            spherical_harmonic(n, m-1, theta, phi)
        second = np.sqrt((n-m) * (n+m+1)) / exp_phi * \
            spherical_harmonic(n, m+1, theta, phi)
        res = (first-second)/2 * (-1)

    return res


def legendre_function(n, m, z, cs_phase=True):
    r"""Legendre function of order n and degree m with argument z.

    .. math::

        P_n^m(z) = (-1)^m(1-z^2)^{m/2}\frac{d^m}{dz^m}P_n{z}

    where the Condon-Shotley phase term $(-1)^m$ is dropped when cs_phase=False
    is used.

    Parameters
    ----------
    n : int
        The order
    m : int
        The degree
    z : ndarray, double
        The argument as an array
    cs_phase : bool, optional
        Whether to use include the Condon-Shotley phase term (-1)^m or not

    Returns
    -------
    legendre : ndarray, double
        The Legendre function. This will return zeros if $|m| > n$.

    Note
    ----
    This is a wrapper for the Legendre function implementation from scipy. The
    scipy implementation uses the Condon-Shotley phase. Therefore, the sign
    needs to be flipped here for uneven degrees when dropping the
    Condon-Shotley phase.

    """
    z = np.atleast_1d(z)

    if np.abs(m) > n:
        legendre = np.zeros(z.shape)
    else:
        legendre = np.zeros(z.shape)
        for idx, arg in zip(count(), z):
            leg, _ = _spspecial.lpmn(m, n, arg)
            if np.mod(m, 2) != 0 and not cs_phase:
                legendre[idx] = -leg[-1, -1]
            else:
                legendre[idx] = leg[-1, -1]
    return legendre


def spherical_harmonic_normalization(n, m, norm='full'):
    r"""The normalization factor for real valued spherical harmonics.

    .. math::

        N_n^m = \sqrt{\frac{2n+1}{4\pi}\frac{(n-m)!}{(n+m)!}}

    Parameters
    ----------
    n : int
        The spherical harmonic order.
    m : int
        The spherical harmonic degree.
    norm : 'full', 'semi', optional
        Normalization to use. Can be either fully normalzied on the sphere or
        semi-normalized.

    Returns
    -------
    norm : double
        The normalization factor.


    """
    if np.abs(m) > n:
        factor = 0.0
    else:
        if norm == 'full':
            z = n+m+1
            factor = _spspecial.poch(z, -2*m)
            factor *= (2*n+1)/(4*np.pi)
            if int(m) != 0:
                factor *= 2
            factor = np.sqrt(factor)
        elif norm == 'semi':
            z = n+m+1
            factor = _spspecial.poch(z, -2*m)
            if int(m) != 0:
                factor *= 2
            factor = np.sqrt(factor)
        else:
            raise ValueError("Unknown normalization.")
    return factor


def spherical_harmonic_derivative_theta_real(n, m, theta, phi):
    r"""The derivative of the real valued spherical harmonics with respect
    to the elevation angle $\theta$.

    Parameters
    ----------
    n : int
        The spherical harmonic order.
    m : int
        The spherical harmonic degree.
    theta : ndarray, double
        The elevation angle
    phi : ndarray, double
        The azimuth angle


    Returns
    -------
    derivative : ndarray, double
        The derivative

    Note
    ----
    This implementation does not include the Condon-Shotley phase term.

    """

    m_abs = np.abs(m)
    if n == 0:
        res = np.zeros(theta.shape, dtype=float)
    else:
        first = (n+m_abs)*(n-m_abs+1) * \
            legendre_function(
                n,
                m_abs-1,
                np.cos(theta),
                cs_phase=False)
        second = legendre_function(
            n,
            m_abs+1,
            np.cos(theta),
            cs_phase=False)
        legendre_diff = 0.5*(first - second)

        N_nm = spherical_harmonic_normalization(n, m_abs)

        if m < 0:
            phi_term = np.sin(m_abs*phi)
        else:
            phi_term = np.cos(m_abs*phi)

        res = N_nm * legendre_diff * phi_term

    return res


def spherical_harmonic_derivative_phi_real(n, m, theta, phi):
    r"""The derivative of the real valued spherical harmonics with respect
    to the azimuth angle $\phi$.

    Parameters
    ----------
    n : int
        The spherical harmonic order.
    m : int
        The spherical harmonic degree.
    theta : ndarray, double
        The elevation angle
    phi : ndarray, double
        The azimuth angle


    Returns
    -------
    derivative : ndarray, double
        The derivative

    Note
    ----
    This implementation does not include the Condon-Shotley phase term.

    """
    m_abs = np.abs(m)
    if m == 0:
        res = np.zeros(theta.shape, dtype=float)
    else:
        legendre = legendre_function(n, m_abs, np.cos(theta), cs_phase=False)
        N_nm = spherical_harmonic_normalization(n, m_abs)

        if m < 0:
            phi_term = np.cos(m_abs*phi) * m_abs
        else:
            phi_term = -np.sin(m_abs*phi) * m_abs

        res = N_nm * legendre * phi_term

    return res


def spherical_harmonic_gradient_phi_real(n, m, theta, phi):
    r"""The gradient of the real valued spherical harmonics with respect
    to the azimuth angle $\phi$.

    Parameters
    ----------
    n : int
        The spherical harmonic order.
    m : int
        The spherical harmonic degree.
    theta : ndarray, double
        The elevation angle
    phi : ndarray, double
        The azimuth angle


    Returns
    -------
    derivative : ndarray, double
        The derivative

    Note
    ----
    This implementation does not include the Condon-Shotley phase term.

    References
    ----------
    .. [1]  J. Du, C. Chen, V. Lesur, and L. Wang, “Non-singular spherical
            harmonic expressions of geomagnetic vector and gradient tensor
            fields in the local north-oriented reference frame,” Geoscientific
            Model Development, vol. 8, no. 7, pp. 1979–1990, Jul. 2015.

    """
    m_abs = np.abs(m)
    if m == 0:
        res = np.zeros(theta.shape, dtype=float)
    else:
        first = (n+m_abs)*(n+m_abs-1) * \
            legendre_function(
                n-1,
                m_abs-1,
                np.cos(theta),
                cs_phase=False)
        second = legendre_function(
            n-1,
            m_abs+1,
            np.cos(theta),
            cs_phase=False)
        legendre_diff = 0.5*(first + second)
        N_nm = spherical_harmonic_normalization(n, m_abs)

        if m < 0:
            phi_term = np.cos(m_abs*phi)
        else:
            phi_term = -np.sin(m_abs*phi)

        res = N_nm * legendre_diff * phi_term

    return res


def legendre_coefficients(order):
    """Calculate the coefficients of a Legendre polynomial of the
    specified order.

    Parameters
    ----------
    order : int
        The order of the polynomial

    Returns
    -------
    coefficients : ndarray, double
        The coefficients of the polynomial

    """
    leg = np.polynomial.legendre.Legendre.basis(order)
    coefficients = np.polynomial.legendre.leg2poly(leg.coef)

    return coefficients


def chebyshev_coefficients(order):
    """Calculate the coefficients of a Chebyshev polynomial of the
    specified order.

    Parameters
    ----------
    order : int
        The order of the polynomial

    Returns
    -------
    coefficients : ndarray, double
        The coefficients of the polynomial

    """
    cheb = np.polynomial.chebyshev.Chebyshev.basis(order)
    coefficients = np.polynomial.chebyshev.cheb2poly(cheb.coef)

    return coefficients
